import abc
import typing as ta

from omlish import dataclasses as dc
from omlish import lang

from ..services import Service
from ..services import ServiceOption
from ..services import ServiceRequest
from ..services import ServiceResponse
from .vectors import Vector


##


@dc.dataclass(frozen=True)
class VectorIndexed(lang.Final):
    v: ta.Any
    vec: Vector


##


VectorIndexInput: ta.TypeAlias = VectorIndexed
VectorIndexNew: ta.TypeAlias = VectorIndexInput
VectorIndexOutput: ta.TypeAlias = None

VectorIndexOptions: ta.TypeAlias = ServiceOption


@dc.dataclass(frozen=True, kw_only=True)
class VectorIndexRequest(
    ServiceRequest[
        VectorIndexInput,
        VectorIndexOptions,
        VectorIndexNew,
    ],
    lang.Final,
):
    pass


@dc.dataclass(frozen=True, kw_only=True)
class VectorIndexResponse(ServiceResponse[VectorIndexOutput], lang.Final):
    pass


class VectorIndexService(
    Service[
        VectorIndexRequest,
        VectorIndexOptions,
        VectorIndexNew,
        VectorIndexResponse,
    ],
    lang.Abstract,
):
    @abc.abstractmethod
    def invoke(self, request: VectorIndexRequest) -> VectorIndexResponse:
        raise NotImplementedError



import abc
import typing as ta

from omlish import dataclasses as dc
from omlish import lang

from .services import Service
from .services import ServiceOption
from .services import ServiceRequest
from .services import ServiceResponse


##


@dc.dataclass(frozen=True, kw_only=True)
class SearchHit(lang.Final):
    title: str | None
    url: str | None
    description: str | None = None
    snippets: lang.SequenceNotStr[str] | None


@dc.dataclass(frozen=True, kw_only=True)
class SearchHits(lang.Final):
    l: ta.Sequence[SearchHit]

    total_results: int | None = None


##


SearchInput: ta.TypeAlias = str
SearchNew: ta.TypeAlias = str
SearchOutput: ta.TypeAlias = SearchHits

SearchOptions: ta.TypeAlias = ServiceOption


@dc.dataclass(frozen=True, kw_only=True)
class SearchRequest(
    ServiceRequest[
        SearchInput,
        SearchOptions,
        SearchNew,
    ],
    lang.Final,
):
    @dc.validate
    def _validate_v(self) -> bool:
        return isinstance(self.v, str)


@dc.dataclass(frozen=True, kw_only=True)
class SearchResponse(ServiceResponse[SearchOutput], lang.Final):
    pass


class SearchService(
    Service[
        SearchRequest,
        SearchOptions,
        SearchNew,
        SearchResponse,
    ],
    lang.Abstract,
):
    @abc.abstractmethod
    def invoke(self, request: SearchRequest) -> SearchResponse:
        raise NotImplementedError



from .chat.formats import (  # noqa
    JSON_RESPONSE_FORMAT,
    JsonResponseFormat,
    ResponseFormat,
    TEXT_RESPONSE_FORMAT,
    TextResponseFormat,
)

from .chat.history import (  # noqa
    ChatHistory,
    ChatHistoryModel,
    ListChatHistory,
)

from .chat.messages import (  # noqa
    AiMessage,
    Chat,
    Message,
    SystemMessage,
    ToolExecResultMessage,
    UserMessage,
)

from .chat.models import (  # noqa
    AiChoice,
    AiChoices,
    ChatInput,
    ChatModel,
    ChatOutput,
    ChatRequest,
    ChatResponse,
)

from .chat.options import (  # noqa
    ChatOption,
    ChatOptions,
)

from .chat.tools import (  # noqa
    Tool,
    ToolDtype,
    ToolExecRequest,
    ToolParam,
    ToolSpec,
)

from .content.content import (  # noqa
    Content,
    ExtendedContent,
)

from .content.images import (  # noqa
    Image,
)

from .content.placeholders import (  # noqa
    Placeholder,
)

from .generative import (  # noqa
    Generative,

    GenerativeOption,
    TopK,
    Temperature,
    MaxTokens,
)

from .models import (  # noqa
    FinishReason,

    TokenUsage,

    Model,
    ModelRequest,
    ModelOption,
    ModelResponse,
)

from .options import (  # noqa
    DuplicateUniqueOptionError,
    Option,
    Options,
    ScalarOption,
    UniqueOption,
)

from .prompts import (  # noqa
    PromptInput,
    PromptModel,
    PromptNew,
    PromptOutput,
    PromptRequest,
    PromptOptions,
    PromptResponse,
)

from .services import (  # noqa
    Service,
    ServiceOption,
    ServiceRequest,
    ServiceResponse,
)

from .vectors.embeddings import (  # noqa
    EmbeddingModel,
    EmbeddingOptions,
    EmbeddingOutput,
    EmbeddingRequest,
    EmbeddingResponse,
)

from .vectors.index import (  # noqa
    VectorIndexed,
    VectorIndexService,
)

from .vectors.search import (  # noqa
    VectorHit,
    VectorHits,
    VectorSearch,
)

from .vectors.similarity import (  # noqa
    CALC_NP_SIMILARITIES_FUNCS,
    Similarity,
    calc_np_cosine_similarities,
    calc_np_dot_similarities,
    calc_np_similarities,
)

from .vectors.vectors import (  # noqa
    Vector,
    Vectorable,
)

from .vectors.stores import (  # noqa
    VectorStore,
)


##


from omlish.lang.imports import _register_conditional_import  # noqa

_register_conditional_import('omlish.marshal', '.marshal', __package__)



"""
FIXME:
 - it outputs newline-terminated so buffer and chop on newlines - DelimitingBuffer again
"""
import ctypes as ct

import llama_cpp
from llama_cpp import _logger as lcl  # noqa

from omlish import lang


##


@llama_cpp.llama_log_callback
def llama_log_callback(
        level: int,
        text: bytes,
        user_data: ct.c_void_p,
) -> None:
    log_level = lcl.GGML_LOG_LEVEL_TO_LOGGING_LEVEL[level] if level != 5 else lcl._last_log_level  # noqa
    if (ts := text.decode('utf-8').rstrip()) != '.' and ts.strip():  # it tends to just output dots lol
        lcl.logger.log(log_level, ts)
    lcl._last_log_level = log_level  # noqa


@lang.cached_function
def install_logging_hook() -> None:
    llama_cpp.llama_log_set(llama_log_callback, ct.c_void_p(0))



import typing as ta

from omlish import lang

from ..search import SearchHit
from ..search import SearchHits
from ..search import SearchRequest
from ..search import SearchResponse
from ..search import SearchService


if ta.TYPE_CHECKING:
    import duckduckgo_search
else:
    duckduckgo_search = lang.proxy_import('duckduckgo_search')


class DuckduckgoSearchService(SearchService):
    def invoke(
            self,
            request: SearchRequest,
    ) -> SearchResponse:
        ddgs = duckduckgo_search.DDGS()
        res = ddgs.text(request.v)
        return SearchResponse(v=SearchHits(
            l=[
                SearchHit(
                    title=d.get('title'),
                    url=d.get('href'),
                    description=d.get('description'),
                    snippets=[body] if (body := d.get('body')) is not None else None,
                )
                for d in res
            ],
        ))



import abc
import typing as ta

from omlish import dataclasses as dc
from omlish import lang

from ..options import Option
from ..services import Service
from ..services import ServiceOption
from ..services import ServiceRequest
from ..services import ServiceResponse
from .similarity import Similarity
from .vectors import Vector


##


class VectorSearchOption(Option, lang.Abstract):
    pass


@dc.dataclass(frozen=True)
class VectorSearchSimilarity(Option, lang.Final):
    similarity: Similarity


##


@dc.dataclass(frozen=True)
class VectorSearch(lang.Final):
    vec: Vector

    _: dc.KW_ONLY

    k: int = 10


@dc.dataclass(frozen=True)
class VectorHit(lang.Final):
    v: ta.Any
    score: float


@dc.dataclass(frozen=True)
class VectorHits(lang.Final):
    l: ta.Sequence[VectorHit]


##


VectorSearchInput: ta.TypeAlias = VectorSearch
VectorSearchNew: ta.TypeAlias = str
VectorSearchOutput: ta.TypeAlias = str

VectorSearchOptions: ta.TypeAlias = ServiceOption | VectorSearchOption


@dc.dataclass(frozen=True, kw_only=True)
class VectorSearchRequest(
    ServiceRequest[
        VectorSearchInput,
        VectorSearchOptions,
        VectorSearchNew,
    ],
    lang.Final,
):
    @dc.validate
    def _validate_v(self) -> bool:
        return isinstance(self.v, str)


@dc.dataclass(frozen=True, kw_only=True)
class VectorSearchResponse(ServiceResponse[VectorSearchOutput], lang.Final):
    pass


class VectorSearchService(
    Service[
        VectorSearchRequest,
        VectorSearchOptions,
        VectorSearchNew,
        VectorSearchResponse,
    ],
    lang.Abstract,
):
    @abc.abstractmethod
    def invoke(self, request: VectorSearchRequest) -> VectorSearchResponse:
        raise NotImplementedError



"""
TODO:
 - metadata?
 - ListOfContent? what to name
"""
import typing as ta

from omlish import lang


##


class ExtendedContent(lang.Abstract, lang.PackageSealed):
    pass


##


SingleContent: ta.TypeAlias = ta.Union[  # noqa
    str,
    ExtendedContent,
]


Content: ta.TypeAlias = ta.Union[  # noqa
    ta.Sequence['Content'],
    SingleContent,
]



from omlish import lang
from omlish import marshal as msh

from .messages import Message


@lang.static_init
def _install_standard_marshalling() -> None:
    msgs_poly = msh.polymorphism_from_subclasses(Message, naming=msh.Naming.SNAKE, strip_suffix=True)
    msh.STANDARD_MARSHALER_FACTORIES[0:0] = [msh.PolymorphismMarshalerFactory(msgs_poly)]
    msh.STANDARD_UNMARSHALER_FACTORIES[0:0] = [msh.PolymorphismUnmarshalerFactory(msgs_poly)]



from omlish import check
from omlish import dataclasses as dc
from omlish import lang
from omlish import marshal as msh
from omlish import reflect as rfl
from omlish.funcs import match as mfs

from .options import Option
from .options import Options
from .options import ScalarOption


##


def _build_option_poly(rty: rfl.Type) -> msh.Polymorphism:
    ty: type[Option] = check.issubclass(check.isinstance(rty, type), Option)
    check.state(lang.is_abstract_class(ty))
    return msh.polymorphism_from_subclasses(
        ty,
        naming=msh.Naming.SNAKE,
    )


class OptionMarshalerFactory(msh.MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, ScalarOption) and not lang.is_abstract_class(rty))  # noqa
    def _build_scalar(self, ctx: msh.MarshalContext, rty: rfl.Type) -> msh.Marshaler:
        dc_rfl = dc.reflect(check.isinstance(rty, type))
        v_rty = check.single(dc_rfl.generic_replaced_field_annotations.values())
        v_m = ctx.make(v_rty)
        return msh.WrappedMarshaler(lambda _, o: o.v, v_m)

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, Option) and lang.is_abstract_class(rty))
    def _build_abstract(self, ctx: msh.MarshalContext, rty: rfl.Type) -> msh.Marshaler:
        return msh.make_polymorphism_marshaler(
            _build_option_poly(rty).impls,
            msh.WrapperTypeTagging(),
            ctx,
        )


class OptionUnmarshalerFactory(msh.UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, ScalarOption) and not lang.is_abstract_class(rty))  # noqa
    def _build_scalar(self, ctx: msh.UnmarshalContext, rty: rfl.Type) -> msh.Unmarshaler:
        rty = check.isinstance(rty, type)
        dc_rfl = dc.reflect(rty)
        v_rty = check.single(dc_rfl.generic_replaced_field_annotations.values())
        v_u = ctx.make(v_rty)
        return msh.WrappedUnmarshaler(lambda _, v: rty(v), v_u)

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, Option) and lang.is_abstract_class(rty))
    def _build_abstract(self, ctx: msh.UnmarshalContext, rty: rfl.Type) -> msh.Unmarshaler:
        return msh.make_polymorphism_unmarshaler(
            _build_option_poly(rty).impls,
            msh.WrapperTypeTagging(),
            ctx,
        )


##


def _build_options_impls(rty: rfl.Type) -> msh.Impls:
    gty = check.isinstance(rty, rfl.Generic)
    check.is_(gty.cls, Options)

    opt_cls_set: set[type[Option]] = {
        check.issubclass(check.isinstance(a, type), Option)
        for a in check.isinstance(check.single(gty.args), rfl.Union).args
    }

    opt_impls: list[msh.Impl] = []
    for opt_cls in opt_cls_set:
        opt_impls.extend(_build_option_poly(opt_cls).impls)

    return msh.Impls(opt_impls)


class OptionsMarshalerFactory(msh.MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and rty.cls is Options)
    def _build(self, ctx: msh.MarshalContext, rty: rfl.Type) -> msh.Marshaler:
        opt_m = msh.make_polymorphism_marshaler(
            msh.Impls(_build_options_impls(rty)),
            msh.WrapperTypeTagging(),
            ctx,
        )
        return msh.IterableMarshaler(opt_m)


class OptionsUnmarshalerFactory(msh.UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and rty.cls is Options)
    def _build(self, ctx: msh.UnmarshalContext, rty: rfl.Type) -> msh.Unmarshaler:
        opt_u = msh.make_polymorphism_unmarshaler(
            msh.Impls(_build_options_impls(rty)),
            msh.WrapperTypeTagging(),
            ctx,
        )
        return msh.IterableUnmarshaler(lambda it: Options(*it), opt_u)  # noqa


##


@lang.static_init
def _install_standard_marshalling() -> None:
    msh.STANDARD_MARSHALER_FACTORIES[0:0] = [
        OptionMarshalerFactory(),
        OptionsMarshalerFactory(),
    ]
    msh.STANDARD_UNMARSHALER_FACTORIES[0:0] = [
        OptionUnmarshalerFactory(),
        OptionsUnmarshalerFactory(),
    ]



import os

from omlish import check
from omlish.formats import json
from omlish.http import all as http
from omlish.secrets.secrets import Secret

from ...prompts import PromptModel
from ...prompts import PromptRequest
from ...prompts import PromptResponse


class OpenaiPromptModel(PromptModel):
    model = 'gpt-3.5-turbo-instruct'

    def __init__(
            self,
            *,
            api_key: Secret | str | None = None,
    ) -> None:
        super().__init__()
        self._api_key = Secret.of(api_key if api_key is not None else os.environ['OPENAI_API_KEY'])

    def invoke(self, t: PromptRequest) -> PromptResponse:
        raw_request = dict(
            model=self.model,
            prompt=t.v,
            temperature=0,
            max_tokens=1024,
            top_p=1,
            frequency_penalty=0.,
            presence_penalty=0.,
            stream=False,
        )

        raw_response = http.request(
            'https://api.openai.com/v1/completions',
            headers={
                http.consts.HEADER_CONTENT_TYPE: http.consts.CONTENT_TYPE_JSON,
                http.consts.HEADER_AUTH: http.consts.format_bearer_auth_header(check.not_none(self._api_key).reveal()),
            },
            data=json.dumps(raw_request).encode('utf-8'),
        )

        response = json.loads(check.not_none(raw_response.data).decode('utf-8'))

        choice = check.single(response['choices'])
        return PromptResponse(v=choice['text'])



import typing as ta

from omlish import dataclasses as dc
from omlish import lang

from ..content.content import Content
from .tools import ToolExecRequest


##


class Message(lang.Abstract, lang.Sealed):
    pass


@dc.dataclass(frozen=True)
class SystemMessage(Message, lang.Final):
    s: str


@dc.dataclass(frozen=True)
class UserMessage(Message, lang.Final):
    c: Content
    name: str | None = dc.xfield(None, repr_fn=dc.opt_repr)


@dc.dataclass(frozen=True)
class AiMessage(Message, lang.Final):
    s: str | None = dc.xfield(None, repr_fn=dc.opt_repr)
    tool_exec_requests: ta.Sequence['ToolExecRequest'] | None = dc.xfield(None, repr_fn=dc.opt_repr)


@dc.dataclass(frozen=True)
class ToolExecResultMessage(Message, lang.Final):
    id: str
    name: str
    s: str


##


Chat: ta.TypeAlias = ta.Sequence[Message]



"""
TODO:
 - Argument

https://mwparserfromhell.readthedocs.io/en/latest/api/mwparserfromhell.nodes.html#module-mwparserfromhell.nodes
"""
import abc
import operator
import typing as ta

from omlish import dataclasses as dc
from omlish import lang
from omlish import marshal as msh


if ta.TYPE_CHECKING:
    import mwparserfromhell as mfh
    import mwparserfromhell.nodes as mfn

else:
    mfh = lang.proxy_import('mwparserfromhell')
    mfn = lang.proxy_import('mwparserfromhell.nodes')


Wikicode: ta.TypeAlias = 'mfh.wikicode.Wikicode'


##


@dc.dataclass(frozen=True)
class Node(abc.ABC):  # noqa
    pass


@dc.dataclass(frozen=True)
class ContentNode(Node, abc.ABC):  # noqa
    pass


Content: ta.TypeAlias = ta.Sequence[ContentNode]


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['body'], omit_if=operator.not_)
class Doc(Node):
    body: Content = ()


@dc.dataclass(frozen=True)
class Text(ContentNode):
    s: str


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['title', 'text'], omit_if=operator.not_)
class WikiLink(ContentNode):
    title: Content = ()
    text: Content = ()


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['title', 'url'], omit_if=operator.not_)
class ExternalLink(ContentNode):
    title: Content = ()
    url: Content = ()


@dc.dataclass(frozen=True)
class Html(ContentNode):
    s: str


@dc.dataclass(frozen=True)
class Comment(ContentNode):
    s: str


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['name', 'value'], omit_if=operator.not_)
class Parameter(Node):
    name: Content = ()
    value: Content = ()


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['name', 'params'], omit_if=operator.not_)
class Template(ContentNode):
    name: Content = ()
    params: ta.Sequence[Parameter] = ()


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['name', 'value'], omit_if=operator.not_)
class Attribute(Node):
    name: Content = ()
    value: Content = ()


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['l', 'atts', 'body', 'r'], omit_if=operator.not_)
class Tag(ContentNode):
    l: Content = ()
    atts: ta.Sequence[Attribute] = ()
    body: Content = ()
    r: Content = ()


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['title'], omit_if=operator.not_)
class Heading(ContentNode):
    title: Content = ()
    level: int = 0


@dc.dataclass(frozen=True)
@msh.update_fields_metadata(['name', 'default'], omit_if=operator.not_)
class Argument(ContentNode):
    name: Content = ()
    default: Content = ()


##


def _install_msh_poly(cls: type) -> None:
    p = msh.polymorphism_from_subclasses(cls, naming=msh.Naming.SNAKE)
    msh.STANDARD_MARSHALER_FACTORIES[0:0] = [msh.PolymorphismMarshalerFactory(p)]
    msh.STANDARD_UNMARSHALER_FACTORIES[0:0] = [msh.PolymorphismUnmarshalerFactory(p)]


_install_msh_poly(Node)
_install_msh_poly(ContentNode)


##


def one_text(nodes: Content) -> str | None:
    if len(nodes) == 1 and isinstance(n := nodes[0], Text):
        return n.s
    return None


_TRAVERSAL_ATTRS_BY_TYPE: ta.Mapping[type[Node], ta.Sequence[str]] = {
    Doc: ['body'],
    Text: [],
    WikiLink: ['title', 'text'],
    ExternalLink: ['title', 'url'],
    Html: [],
    Comment: [],
    Parameter: ['name', 'value'],
    Template: ['name', 'params'],
    Attribute: ['name', 'value'],
    Tag: ['l', 'atts', 'body', 'r'],
    Heading: ['title'],
    Argument: ['name', 'default'],
}


def traverse_node(root: Node) -> ta.Iterator[tuple[ta.Sequence[tuple[Node, str]], Node]]:
    path: list[tuple[Node, str]] = []

    def rec(n: Node) -> ta.Iterator[tuple[ta.Sequence[tuple[Node, str]], Node]]:
        yield (path, n)

        for a in _TRAVERSAL_ATTRS_BY_TYPE[n.__class__]:
            if not (v := getattr(n, a)):
                continue

            path.append((n, a))
            for c in v:
                yield from rec(c)
            path.pop()

    yield from rec(root)


##


class NodeBuilder:
    def build_parameter(self, n: 'mfn.extras.Parameter') -> Parameter:
        if not isinstance(n, mfn.extras.Parameter):
            raise TypeError(n)

        return Parameter(
            self.build_content(n.name),
            self.build_content(n.value),
        )

    def build_attribute(self, n: 'mfn.extras.Attribute') -> Attribute:
        if not isinstance(n, mfn.extras.Attribute):
            raise TypeError(n)

        return Attribute(
            self.build_content(n.name),
            self.build_content(n.value),
        )

    def build_content_node(self, n: 'mfn.Node') -> ContentNode:
        if isinstance(n, mfn.Comment):
            return Comment(n.contents)

        elif isinstance(n, mfn.Text):
            return Text(n.value)

        elif isinstance(n, mfn.Template):
            return Template(
                self.build_content(n.name),
                list(map(self.build_parameter, n.params)),
            )

        elif isinstance(n, mfn.Wikilink):
            return WikiLink(
                self.build_content(n.title),
                self.build_content(n.text),
            )

        elif isinstance(n, mfn.ExternalLink):
            return ExternalLink(
                self.build_content(n.title),
                self.build_content(n.url),
            )

        elif isinstance(n, mfn.HTMLEntity):
            return Html(n.value)

        elif isinstance(n, mfn.Tag):
            return Tag(
                self.build_content(n.tag),
                list(map(self.build_attribute, n.attributes)),
                self.build_content(n.contents),
                self.build_content(n.closing_tag),
            )

        elif isinstance(n, mfn.Heading):
            return Heading(
                self.build_content(n.title),
                n.level,
            )

        elif isinstance(n, mfn.Argument):
            return Argument(
                self.build_content(n.name),
                self.build_content(n.default),
            )

        else:
            raise TypeError(n)

    def build_content(self, w: Wikicode | ta.Iterable['mfn.Node'] | None) -> Content:
        if w is None:
            return ()

        elif isinstance(w, mfh.wikicode.Wikicode):
            return [self.build_content_node(c) for c in w.nodes]

        elif isinstance(w, ta.Iterable):
            return [self.build_content_node(c) for c in w]

        else:
            raise TypeError(w)


##


def parse(
        value: ta.Any,
        context: int = 0,
        *,
        skip_style_tags: bool = False,
        **kwargs: ta.Any,
) -> Wikicode:
    return mfh.parse(
        value,
        context,
        skip_style_tags=skip_style_tags,
        **kwargs,
    )


def parse_doc(s: str) -> Doc:
    wiki = parse(s)
    content = NodeBuilder().build_content(wiki)
    return Doc(content)



import typing as ta

from omlish import check
from omlish import lang


if ta.TYPE_CHECKING:
    import huggingface_hub as hf
    import huggingface_hub.errors  # noqa
    import huggingface_hub.utils  # noqa
else:
    hf = lang.proxy_import('huggingface_hub', extras=[
        'errors',
        'utils',
    ])


def is_repo_cached(
        id: str,  # noqa
        *,
        revisions: ta.Iterable[str] | None = None,
        cache_dir: str | None = None,
) -> bool:
    check.not_isinstance(revisions, str)

    try:
        cache_info = hf.utils.scan_cache_dir(cache_dir=cache_dir)
    except hf.errors.CacheNotFound:
        return False

    revision_set = frozenset(revisions) if revisions is not None else None

    for repo in cache_info.repos:
        if id != repo.repo_id:
            continue

        if revision_set is not None and not any(rev in revision_set for rev in repo.revisions):
            continue

        return True

    return False



"""
https://github.com/langchain4j/langchain4j/tree/d069798d52430203084dcb2114cbbd3afc3ff6cb/langchain4j-core/src/main/java/dev/langchain4j

==

agent
  tool
    JsonSchemaProperty
    P
    Tool
    ToolExecutionRequest
    ToolMemoryId
    ToolParameters
    ToolSpecification
    ToolSpecifications

code
  CodeExecutionEngine

data
  audio
    Audio

  document
    BlankDocumentException
    Document
    DocumentLoader
    DocumentParser
    DocumentSource
    DocumentSplitter
    DocumentTransformer
    Metadata

  embedding
    Embedding

  image
    Image

  message
    AiMessage
    AudioContent
    ChatMessage
    ChatMessageDeserializer
    ChatMessageJsonCodec
    ChatMessageSerializer
    ChatMessageType
    Content
    ContentType
    GsonChatMessageAdapter
    GsonChatMessageJsonCodec
    GsonContentAdapter
    ImageContent
    PdfFileContent
    SystemMessage
    TextContent
    TextFileContent
    ToolExecutionResultMessage
    UserMessage
    VideoContent

  pdf
    PdfFile

  segment
    TextSegment
    TextSegmentTransformer

  text
    TextFile

  video
    Video

internal
  CustomMimeTypesFileTypeDetector
  Exceptions
  GsonJsonCodec
  JacocoIgnoreCoverageGenerated
  Json
  RetryUtils
  TypeUtils
  Utils
  ValidationUtils

memory
  ChatMemory

model
  chat
    listener
      ChatModelErrorContext
      ChatModelListener
      ChatModelRequest
      ChatModelRequestContext
      ChatModelResponse
      ChatModelResponseContext

    request
      json
        JsonArraySchema
        JsonBooleanSchema
        JsonEnumSchema
        JsonIntegerSchema
        JsonNumberSchema
        JsonObjectSchema
        JsonSchema
        JsonSchemaElement
        JsonStringSchema

      ChatRequest
      ResponseFormat
      ResponseFormatType

    response
      ChatResponse

    Capability
    ChatLanguageModel
    DisabledChatLanguageModel
    DisabledStreamingChatLanguageModel
    StreamingChatLanguageModel
    TokenCountEstimator

  embedding
    DimensionAwareEmbeddingModel
    DisabledEmbeddingModel
    EmbeddingModel
    TokenCountEstimator

  image
    DisabledImageModel
    ImageModel

  input
    structured
      DefaultStructuredPromptFactory
      StructuredPrompt
      StructuredPromptProcessor

    DefaultPromptTemplateFactory
    Prompt
    PromptTemplate

  language
    DisabledLanguageModel
    DisabledStreamingLanguageModel
    LanguageModel
    StreamingLanguageModel
    TokenCountEstimator

  moderation
    DisabledModerationModel
    Moderation
    ModerationModel

  output
    structured
      Description

    FinishReason
    Response
    TokenUsage

  scoring
    ScoringModel

  LambdaStreamingResponseHandler
  ModelDisabledException
  StreamingResponseHandler
  Tokenizer

rag
  content
    aggregator
      ContentAggregator
      DefaultContentAggregator
      ReRankingContentAggregator
      ReciprocalRankFuser

    injector
      ContentInjector
      DefaultContentInjector

    retriever
      ContentRetriever
      EmbeddingStoreContentRetriever
      WebSearchContentRetriever

    Content

  query
    router
      DefaultQueryRouter
      LanguageModelQueryRouter
      QueryRouter

    transformer
      CompressingQueryTransformer
      DefaultQueryTransformer
      ExpandingQueryTransformer
      QueryTransformer

    Metadata
    Query

  AugmentationRequest
  AugmentationResult
  DefaultRetrievalAugmentor
  RetrievalAugmentor

retriever
  EmbeddingStoreRetriever
  Retriever

spi
  ServiceHelper
  data
    document
      parser
        DocumentParserFactory
      splitter
        DocumentSplitterFactory
    message
      ChatMessageJsonCodecFactory
  json
    JsonCodecFactory
  model
    embedding
      EmbeddingModelFactory
  prompt
    PromptTemplateFactory
    structured
      StructuredPromptFactory

store
  embedding
    filter
      comparison
        IsEqualTo
        IsGreaterThan
        IsGreaterThanOrEqualTo
        IsIn
        IsLessThan
        IsLessThanOrEqualTo
        IsNotEqualTo
        IsNotIn
        NumberComparator
        TypeChecker
        UUIDComparator

      logical
        And
        Not
        Or

      Filter
      FilterParser
      MetadataFilterBuilder

    CosineSimilarity
    EmbeddingMatch
    EmbeddingSearchRequest
    EmbeddingSearchResult
    EmbeddingStore
    EmbeddingStoreIngestor
    RelevanceScore

  memory
    chat
      ChatMemoryStore
      InMemoryChatMemoryStore

web
  search
    WebSearchEngine
    WebSearchInformationResult
    WebSearchOrganicResult
    WebSearchRequest
    WebSearchResults
    WebSearchTool
"""



"""
TODO:
 - udf name manager - acquire leases to unique names, release back to pool, replace with dummy
"""
import sqlite3

from omlish import lang

from ..vectors.index import VectorIndexed
from ..vectors.search import VectorHit
from ..vectors.search import VectorHits
from ..vectors.search import VectorSearch
from ..vectors.similarity import CALC_NP_SIMILARITIES_FUNCS
from ..vectors.similarity import Similarity
from ..vectors.stores import VectorStore
from ..vectors.vectors import Vector


class SqliteVectorStore(VectorStore):

    def __init__(
            self,
            db: sqlite3.Connection,
            table_name: str,
    ) -> None:
        super().__init__()

        self._db = db
        self._table_name = table_name

        db.execute(''.join([
            f'create table if not exists {self._table_name} (',
            ', ' .join([
                'v integer primary key',
                f'vec blob',
            ]),
            ')',
        ]))

    def index(self, doc: VectorIndexed) -> None:
        self._db.execute(
            f'insert into {self._table_name} (v, vec) values (?, ?)',  # noqa
            (doc.v, doc.vec.bytes()),
        )

    def search(
            self,
            search: VectorSearch,
            *,
            similarity: Similarity | None = None,
    ) -> VectorHits:
        snp = search.vec.np()
        npfn = CALC_NP_SIMILARITIES_FUNCS[similarity or Similarity.DOT]

        def calc_score(binary):
            rnp = Vector(binary).np()
            score = float(npfn(rnp.reshape(1, -1), snp)[0])
            return score

        udf_name = 'calc_score'
        self._db.create_function(udf_name, 1, calc_score)
        try:
            rows = self._db.execute(
                f'select v, calc_score(vec) as score '  # noqa
                f'from {self._table_name} '
                f'order by score desc '
                f'limit {search.k}',
            )

            ret = []
            for row in rows:
                ret.append(VectorHit(row[0], row[1]))
            return VectorHits(ret)

        finally:
            self._db.create_function(udf_name, 0, lang.void)



import enum

from omlish import dataclasses as dc
from omlish import lang

from ..vectors.vectors import Vector


@dc.dataclass(frozen=True)
class Dtype(lang.Final):
    name: str
    cls: type

    _: dc.KW_ONLY

    primitive: bool = False


class Dtypes(enum.Enum):
    STR = Dtype('str', str, primitive=True)
    BYTES = Dtype('bytes', bytes, primitive=True)

    INT = Dtype('int', int, primitive=True)
    FLOAT = Dtype('float', float, primitive=True)

    VECTOR = Dtype('vector', Vector)



import abc

from omlish import lang

from .index import VectorIndexed
from .search import VectorHits
from .search import VectorSearch
from .similarity import Similarity


class VectorStore(lang.Abstract):
    @abc.abstractmethod
    def index(self, doc: VectorIndexed) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def search(
            self,
            search: VectorSearch,
            *,
            similarity: Similarity | None = None,
    ) -> VectorHits:
        raise NotImplementedError



import collections.abc
import dataclasses as dc
import typing as ta

from omlish import lang
from omlish import marshal as msh
from omlish import reflect as rfl
from omlish.funcs import match as mfs

from .content import Content
from .content import ExtendedContent
from .content import SingleContent
from .images import Image
from .placeholders import Placeholder  # noqa


##


# TODO: This hack should be obsolete with 3.14 / PEP 649, but reflect needs to grow recursive type support.
class MarshalContent(lang.NotInstantiable, lang.Final):
    pass


MarshalContentUnion: ta.TypeAlias = ta.Union[  # noqa
    ta.Sequence[MarshalContent],
    SingleContent,
]


_MARSHAL_CONTENT_UNION_RTY = rfl.type_(MarshalContentUnion)


@dc.dataclass(frozen=True)
class _ContentMarshaler(msh.Marshaler):
    et: msh.Marshaler

    def marshal(self, ctx: msh.MarshalContext, o: ta.Any) -> msh.Value:
        if isinstance(o, str):
            return o
        elif isinstance(o, ta.Sequence):
            return [self.marshal(ctx, e) for e in o]
        elif isinstance(o, ExtendedContent):
            return self.et.marshal(ctx, o)
        else:
            raise TypeError(o)


class _ContentMarshalerFactory(msh.MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: rty is MarshalContent or rty == _MARSHAL_CONTENT_UNION_RTY)
    def _build(self, ctx: msh.MarshalContext, rty: rfl.Type) -> msh.Marshaler:
        return _ContentMarshaler(ctx.make(ExtendedContent))


@dc.dataclass(frozen=True)
class _ContentUnmarshaler(msh.Unmarshaler):
    et: msh.Unmarshaler

    def unmarshal(self, ctx: msh.UnmarshalContext, v: msh.Value) -> ta.Any:
        if isinstance(v, str):
            return v
        elif isinstance(v, ta.Sequence):
            return [self.unmarshal(ctx, e) for e in v]
        elif isinstance(v, collections.abc.Mapping):
            return self.et.unmarshal(ctx, v)  # noqa
        else:
            raise TypeError(v)


class _ContentUnmarshalerFactory(msh.UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: rty is MarshalContent or rty == _MARSHAL_CONTENT_UNION_RTY)
    def _build(self, ctx: msh.UnmarshalContext, rty: rfl.Type) -> msh.Unmarshaler:
        return _ContentUnmarshaler(ctx.make(ExtendedContent))


##


class _ImageMarshaler(msh.Marshaler):
    def marshal(self, ctx: msh.MarshalContext, o: ta.Any) -> msh.Value:
        raise NotImplementedError


class _ImageUnmarshaler(msh.Unmarshaler):
    def unmarshal(self, ctx: msh.UnmarshalContext, v: msh.Value) -> ta.Any:
        raise NotImplementedError


##


@lang.static_init
def _install_standard_marshalling() -> None:
    extended_content_poly = msh.polymorphism_from_subclasses(ExtendedContent, naming=msh.Naming.SNAKE)
    msh.STANDARD_MARSHALER_FACTORIES[0:0] = [
        msh.PolymorphismMarshalerFactory(extended_content_poly),
        msh.TypeMapMarshalerFactory({Image: _ImageMarshaler()}),
        _ContentMarshalerFactory(),
    ]
    msh.STANDARD_UNMARSHALER_FACTORIES[0:0] = [
        msh.PolymorphismUnmarshalerFactory(extended_content_poly),
        msh.TypeMapUnmarshalerFactory({Image: _ImageUnmarshaler()}),
        _ContentUnmarshalerFactory(),
    ]

    msh.GLOBAL_REGISTRY.register(
        Content,
        msh.ReflectOverride(MarshalContent),
        identity=True,
    )



import typing as ta

from omlish import check
from omlish import dataclasses as dc
from omlish import lang


T = ta.TypeVar('T')
OptionT = ta.TypeVar('OptionT', bound='Option')
OptionU = ta.TypeVar('OptionU', bound='Option')
UniqueOptionU = ta.TypeVar('UniqueOptionU', bound='UniqueOption')


##


class Option(lang.Abstract):
    pass


class UniqueOption(Option, lang.Abstract):
    unique_option_cls: ta.ClassVar[type[Option]]

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        if UniqueOption in cls.__bases__:
            try:
                cls.unique_option_cls  # noqa
            except AttributeError:
                cls.unique_option_cls = cls
            else:
                raise TypeError(f'Class already has unique_option_cls: {cls}')


@dc.dataclass(frozen=True)
@dc.extra_params(generic_init=True)
class ScalarOption(Option, lang.Abstract, ta.Generic[T]):
    v: T

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.v!r})'


##


@dc.dataclass()
class DuplicateUniqueOptionError(Exception, ta.Generic[UniqueOptionU]):
    cls: type[UniqueOptionU]
    new: UniqueOptionU
    old: UniqueOptionU


class Options(lang.Final, ta.Generic[OptionT]):
    def __init__(self, *options: OptionT, override: bool = False) -> None:
        super().__init__()

        tmp: list = []
        udct: dict = {}
        for o in options:
            if isinstance(o, UniqueOption):
                uoc = o.unique_option_cls
                if not override:
                    try:
                        exu = udct[uoc]
                    except KeyError:
                        pass
                    else:
                        raise DuplicateUniqueOptionError(uoc, o, check.single(exu))
                ulst = udct.setdefault(uoc, [])
                ulst.append(o)
                tmp.append((o, len(ulst)))
            elif isinstance(o, Option):
                tmp.append(o)
            else:
                raise TypeError(o)

        lst: list = []
        dct: dict = {}
        for o in tmp:
            if isinstance(o, tuple):
                uo, idx = o  # type: ignore
                ulst = udct[uo.unique_option_cls]
                if idx == len(ulst):
                    lst.append(uo)
                    dct[uo.unique_option_cls] = uo
                    dct[type(uo)] = uo
            else:
                lst.append(o)
                dct.setdefault(type(o), []).append(o)

        self._lst = lst
        self._dct = dct

    def without(self, *tys: type) -> ta.Iterator[OptionT]:
        for o in self._lst:
            if not isinstance(o, tys):
                yield o

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({", ".join(map(repr, self._lst))})'

    def __iter__(self) -> ta.Iterator[OptionT]:
        return iter(self._lst)

    def __len__(self) -> int:
        return len(self._lst)

    def __contains__(self, cls: type[OptionU]) -> bool:
        return cls in self._dct

    def __bool__(self) -> bool:
        return bool(self._lst)

    @ta.overload
    def __getitem__(self, idx: int) -> OptionT:
        ...

    @ta.overload
    def __getitem__(self, cls: type[UniqueOptionU]) -> UniqueOptionU:  # type: ignore[overload-overlap]
        ...

    @ta.overload
    def __getitem__(self, cls: type[OptionU]) -> ta.Sequence[OptionU]:
        ...

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._lst[key]
        else:
            return self._dct[key]



"""
TODO:
 - BZ2File / _compression.DecompressionReader does this, but point is parallel decomp
  - chop stream
   - possible at all? pbzip2 explodes
   - https://github.com/ruanhuabin/pbzip2
   - https://en.wikipedia.org/wiki/Bzip2
"""
import abc
import contextlib
import os
import subprocess
import sys
import time
import typing as ta

from omlish import check
from omlish import lang


if ta.TYPE_CHECKING:
    import bz2

    import lz4.frame as lz4_frame

else:
    bz2 = lang.proxy_import('bz2')

    lz4_frame = lang.proxy_import('lz4.frame')


##


class BytesReaderWrapper(ta.IO[bytes], abc.ABC):
    def __init__(self, f: ta.IO[bytes]) -> None:
        super().__init__()
        self._f = f

    def close(self):
        raise TypeError

    def fileno(self):
        return self._f.fileno()

    def flush(self):
        raise TypeError

    def isatty(self):
        return self._f.isatty()

    @abc.abstractmethod
    def read(self, n=-1):
        raise NotImplementedError

    def readable(self):
        return self._f.readable()

    def readline(self, limit=-1):
        raise TypeError

    def readlines(self, hint=-1):
        raise TypeError

    def seek(self, offset, whence=0):
        raise TypeError

    def seekable(self):
        return False

    def tell(self):
        return self._f.tell()

    def truncate(self, size=None):
        raise TypeError

    def writable(self):
        return self._f.writable()

    def write(self, s):
        raise TypeError

    def writelines(self, lines):
        raise TypeError

    def __next__(self):
        raise TypeError

    def __iter__(self):
        raise TypeError

    def __enter__(self):
        raise TypeError

    def __exit__(self, et, e, tb):
        raise TypeError


##


class FileProgressReporter:
    def __init__(
            self,
            f: ta.Any,
            *,
            out: ta.Any = sys.stderr,
            time_interval: float | None = 1.,
            bytes_interval: int | None = None,
    ) -> None:
        super().__init__()
        self._f = f
        self._out = out
        self._time_interval = time_interval
        self._bytes_interval = bytes_interval

        self._bytes_tot = os.fstat(f.fileno()).st_size
        self._bytes_last = 0
        self._time_last = time.time()

    def _time_pos(self) -> tuple[float, float]:
        time_cur = time.time()
        time_ela = time_cur - self._time_last
        return time_cur, time_ela

    def _bytes_pos(self) -> tuple[int, int]:
        bytes_cur = self._f.tell()
        bytes_ela = bytes_cur - self._bytes_last
        return bytes_cur, bytes_ela

    def report(self) -> bool:
        should_report = False

        if self._time_interval is not None:
            time_cur, time_ela = self._time_pos()
            should_report |= time_ela >= self._time_interval
        else:
            time_cur, time_ela = None, None

        if self._bytes_interval is not None:
            bytes_cur, bytes_ela = self._bytes_pos()
            should_report |= bytes_ela >= self._bytes_interval
        else:
            bytes_cur, bytes_ela = None, None

        if not should_report:
            return False

        if time_cur is None:
            time_cur, time_ela = self._time_pos()
        if bytes_cur is None:
            bytes_cur, bytes_ela = self._bytes_pos()

        print(
            f'{bytes_cur:_} B / {self._bytes_tot:_} B - '
            f'{bytes_cur / self._bytes_tot * 100.:.2f} % - '
            f'{int(bytes_ela / time_ela):_} B/s',  # type: ignore
            file=self._out,
        )

        self._bytes_last = bytes_cur
        self._time_last = time_cur

        return True


##


@contextlib.contextmanager
def open_compressed_reader(
        fp: str,
        *,
        use_subprocess: bool = False,
) -> ta.Iterator[tuple[ta.IO, FileProgressReporter | None]]:
    bs: ta.IO
    fpr: FileProgressReporter | None
    with contextlib.ExitStack() as es:
        if fp.endswith('.bz2'):
            if not use_subprocess:
                f = es.enter_context(open(fp, 'rb'))
                fpr = FileProgressReporter(f, time_interval=5)

                bs = es.enter_context(contextlib.closing(bz2.open(f, 'rb')))

            else:
                f = es.enter_context(open(fp, 'rb'))

                # FIXME: os.dup?
                # fpr = FileProgressReporter(f, time_interval=5)
                fpr = None

                proc = subprocess.Popen(['bzip2', '-cdk', fp], stdout=subprocess.PIPE, stdin=f)
                bs = check.not_none(proc.stdout)

        elif fp.endswith('.lz4'):
            if not use_subprocess:
                f = es.enter_context(open(fp, 'rb'))
                fpr = FileProgressReporter(f, time_interval=5)

                bs = es.enter_context(contextlib.closing(lz4_frame.open(f, 'rb')))

            else:
                f = es.enter_context(open(fp, 'rb'))

                # FIXME: os.dup?
                # fpr = FileProgressReporter(f, time_interval=5)
                fpr = None

                proc = subprocess.Popen(['lz4', '-cdk', fp], stdout=subprocess.PIPE, stdin=f)
                bs = check.not_none(proc.stdout)

        else:
            raise RuntimeError(fp)

        yield bs, fpr


##


class MultiFileWriter:
    @ta.runtime_checkable
    class File(ta.Protocol):
        def close(self) -> None:
            ...

        def tell(self) -> int:
            ...

        def write(self, buf: bytes) -> None:
            ...

    def __init__(
            self,
            file_opener: ta.Callable[[str], File],
            file_pat: str,
            file_size: int = 2 * 1024 * 1024 * 1024,
            *,
            use_input_size: bool = False,
    ) -> None:
        super().__init__()

        self._file_opener = file_opener
        self._file_pat = file_pat
        self._file_size = file_size
        self._use_input_size = use_input_size

        self._cur_n = 0
        self._total_b = 0
        self._cur_b = 0
        self._cur_f: MultiFileWriter.File | None = None

    def close(self) -> None:
        if self._cur_f is not None:
            self._cur_f.close()
            self._cur_f = None

        self._cur_b = 0

    def write(self, *bufs: bytes) -> None:
        if self._cur_f is None:
            self._cur_f = self._file_opener(self._file_pat % (self._cur_n,))

        for buf in bufs:
            self._cur_b += len(buf)
            self._total_b += len(buf)
            self._cur_f.write(buf)

        if (self._cur_f.tell() if self._use_input_size else self._cur_b) >= self._file_size:
            self.close()
            self._cur_n += 1


@lang.protocol_check(MultiFileWriter.File)
class Lz4MfwFile:
    def __init__(self, fp: str) -> None:
        super().__init__()
        self._raw_f = open(fp, 'wb')  # noqa
        self._z_f = lz4_frame.open(self._raw_f, 'wb')

    def close(self) -> None:
        self._z_f.close()
        self._raw_f.close()

    def tell(self) -> int:
        return self._raw_f.tell()

    def write(self, buf: bytes) -> None:
        self._z_f.write(buf)



"""
TODO:
 - move omit magic to omdev lol
"""
import abc
import concurrent.futures as cf
import dataclasses as dc
import os.path
import re
import typing as ta
import urllib.request

from omdev.git.magic import GIT_DIFF_OMIT_MAGIC_COMMENT
from omdev.home.secrets import load_secrets
from omdev.tools.git.messages import GitMessageGenerator
from omlish import check
from omlish import lang
from omlish.configs.classes import Configurable
from omlish.subprocesses.sync import subprocesses

from .. import minichain as mc
from ..minichain.backends.mlxlm import MlxlmChatModel
from ..minichain.backends.openai import OpenaiChatModel
from ..minichain.generative import MaxTokens
from ..server.client import McServerClient


GitAiBackendConfigT = ta.TypeVar('GitAiBackendConfigT', bound='GitAiBackend.Config')


##


class GitAiBackend(Configurable[GitAiBackendConfigT], lang.Abstract):
    @dc.dataclass(frozen=True)
    class Config(Configurable.Config):
        max_tokens: int | None = 128

    @abc.abstractmethod
    def run_prompt(self, prompt: str) -> str:
        raise NotImplementedError


#


class OpenaiGitAiBackend(GitAiBackend['OpenaiGitAiBackend.Config']):
    @dc.dataclass(frozen=True)
    class Config(GitAiBackend.Config):
        pass

    def __init__(self, config: Config = Config()) -> None:
        super().__init__(config)

    def run_prompt(self, prompt: str) -> str:
        # FIXME:
        for key in ['OPENAI_API_KEY']:
            if (sec := load_secrets().try_get(key.lower())) is not None:
                os.environ[key] = sec.reveal()

        llm = OpenaiChatModel()

        resp = llm(
            [mc.UserMessage(prompt)],
            *((MaxTokens(self._config.max_tokens),) if self._config.max_tokens is not None else ()),
        )
        return check.non_empty_str(resp.v[0].m.s)


#


def _strip_markdown_code_block(text: str) -> str:
    if (match := re.fullmatch(r'```(?:\w+)?\n(.+?)\n```', text, re.DOTALL)) is None:
        return text
    return match.group(1)


class MlxlmGitAiBackend(GitAiBackend['MlxlmGitAiBackend.Config']):
    @dc.dataclass(frozen=True)
    class Config(GitAiBackend.Config):
        model: str = 'mlx-community/Qwen2.5-Coder-32B-Instruct-8bit'

        run_in_subprocess: bool = True
        subprocess_timeout: float | None = 60.

    def __init__(self, config: Config = Config()) -> None:
        super().__init__(config)

    def _run_prompt(self, prompt: str) -> str:
        llm = MlxlmChatModel(self._config.model)

        resp = llm(
            [mc.UserMessage(prompt)],
            *((MaxTokens(self._config.max_tokens),) if self._config.max_tokens is not None else ()),
        )
        text = check.non_empty_str(resp.v[0].m.s)

        text = _strip_markdown_code_block(text)

        return text

    def run_prompt(self, prompt: str) -> str:
        if self._config.run_in_subprocess:
            # tokenizers installs a pthread_atfork callback at *import* time:
            #   https://github.com/huggingface/tokenizers/blob/4f1a810aa258d287e6936315e63fbf58bde2a980/bindings/python/src/lib.rs#L57
            # then complains about `TOKENIZERS_PARALLELISM` at the next fork (which presumably will happen immediately
            # after this to `git commit`, despite it being conceptually just a fork/exe).
            # TODO: a generic subprocessing minichain service wrapper?
            with cf.ProcessPoolExecutor() as exe:
                return exe.submit(
                    self._run_prompt,
                    prompt,
                ).result(timeout=self._config.subprocess_timeout)

        else:
            return self._run_prompt(prompt)


#


class LocalhostHttpPostGitAiBackend(GitAiBackend['LocalhostHttpPostGitAiBackend.Config']):
    @dc.dataclass(frozen=True)
    class Config(GitAiBackend.Config):
        port: int = 5067
        path: str = ''

    def __init__(self, config: Config = Config()) -> None:
        super().__init__(config)

    @property
    def config(self) -> Config:
        return self._config

    def run_prompt(self, prompt: str) -> str:
        with urllib.request.urlopen(urllib.request.Request(
                f'http://localhost:{self._config.port}/{self._config.path}',
                data=prompt.encode('utf-8'),
        )) as resp:
            return resp.read().decode('utf-8')


#


class McServerGitAiBackend(GitAiBackend['McServerGitAiBackend.Config']):
    @dc.dataclass(frozen=True)
    class Config(GitAiBackend.Config):
        pass

    def __init__(self, config: Config = Config()) -> None:
        super().__init__(config)

    def run_prompt(self, prompt: str) -> str:
        text = McServerClient().prompt(
            prompt,
            launch=True,
        )

        text = _strip_markdown_code_block(text)

        return text


##


# @omlish-manifest omdev.tools.git.messages.GitMessageGeneratorManifest(name='ai')
class AiGitMessageGenerator(GitMessageGenerator):
    def __init__(
            self,
            *,
            backend: GitAiBackend | None = None,
    ) -> None:
        super().__init__()

        if backend is None:
            backend = self.DEFAULT_BACKEND
        self._backend = backend

    DEFAULT_BACKEND: ta.ClassVar[GitAiBackend] = (
        OpenaiGitAiBackend()
        # LocalhostHttpPostGitAiBackend()
        # MlxlmGitAiBackend()
        # McServerGitAiBackend()
    )

    def _should_exclude_file_name(self, fn: str) -> bool:
        if not fn.endswith('.py'):
            return False

        try:
            with open(fn) as f:
                f_src = f.read()
        except FileNotFoundError:
            return False

        return any(
            l.startswith(GIT_DIFF_OMIT_MAGIC_COMMENT)
            for l in f_src.splitlines()
        )

    # TODO: configurable
    DEFAULT_EXCLUDES: ta.ClassVar[ta.Sequence[str]] = [
        '**/.manifests.json',
        '**/_antlr/*',
    ]

    def generate_commit_message(
            self,
            args: GitMessageGenerator.GenerateCommitMessageArgs,
    ) -> GitMessageGenerator.GenerateCommitMessageResult:
        diff_files = subprocesses.check_output(
            'git',
            'diff',
            '--name-only',
            'HEAD',
        ).decode().splitlines()

        excludes = [
            fn
            for fn in diff_files
            if self._should_exclude_file_name(fn)
        ]

        diff = subprocesses.check_output(
            'git',
            'diff',
            'HEAD',
            '--',
            *[f':(exclude){x}' for x in [*self.DEFAULT_EXCLUDES, *excludes]],
            cwd=args.cwd,
        ).decode()

        prompt = '\n\n'.join([
            'Write a short git commit message for the following git diff:',
            f'```\n{diff}\n```',
            '\n'.join([
                'Follow these rules:',
                '- Use imperative tense (e.g., "Fix bug", "Add feature").',
                '- Be concise but descriptive. Avoid vague messages like "Update X".',
                '- Do not mention files explicitly unless necessary.',
                '- If the change is complex, add a second paragraph with more details.',
                'Output only the commit message, with no additional text.',
            ]),
        ])

        msg = self._backend.run_prompt(prompt)

        return GitMessageGenerator.GenerateCommitMessageResult(
            msg=msg,
            confirm=True,
        )



import typing as ta

from omlish import lang


# fmt: off
if ta.TYPE_CHECKING:
    import tinygrad as tg  # noqa
    from tinygrad import nn  # noqa
else:
    tg = lang.proxy_import('tinygrad')
    nn = lang.proxy_import('tinygrad.nn')
# fmt: on


# mypy workaround
# Tensor = tg.Tensor  # type: ignore
# TinyJit = tg.TinyJit  # type: ignore



"""
https://www.mediawiki.org/wiki/Help:Export#Export_format

https://github.com/5j9/wikitextparser
https://github.com/tatuylonen/wikitextprocessor
https://github.com/earwig/mwparserfromhell

https://github.com/WillKoehrsen/wikipedia-data-science/blob/master/notebooks/Downloading%20and%20Parsing%20Wikipedia%20Articles.ipynb

https://en.wikipedia.org/wiki/Help:Wikitext
https://en.wikipedia.org/wiki/Wikipedia:Lua

==

if rev.text and '#invoke' in rev.text.text:
    import wikitextparser as wtp
    parsed = wtp.parse(rev.text.text)
    print(parsed)

    from wikitextprocessor import Wtp
    ctx = Wtp()
    ctx.start_page(page.title)
    tree = ctx.parse(rev.text.text)
    print(tree)

"""
import argparse
import concurrent.futures as cf
import contextlib
import dataclasses as dc
import glob
import io
import logging
import os.path
import signal
import threading
import time
import typing as ta

import lz4.frame
import sqlalchemy as sa
import sqlalchemy.exc

from omlish import cached
from omlish import concurrent as cfu
from omlish import lang
from omlish import marshal as msh
from omlish import multiprocessing as mpu
from omlish.formats import json
from omlish.logs import all as logs
from omlish.os import deathpacts

from . import models as mdl
from .text import mfh  # noqa
from .text import wtp  # noqa
from .utils.progress import ProgressReporter


log = logging.getLogger(__name__)


##


LZ4_JSONL_DIR = os.path.expanduser('~/data/enwiki/json/')


meta = sa.MetaData()
pages_table = sa.Table(
    'pages',
    meta,
    sa.Column('id', sa.Integer(), primary_key=True),
    sa.Column('title', sa.String()),
    sa.Column('page', sa.BLOB),
    sa.Column('doc', sa.BLOB),
    sa.Index('pages_by_title', 'title'),
)


class FileAnalyzer:
    @dc.dataclass(frozen=True, kw_only=True)
    class Context:
        deathpact: deathpacts.Deathpact
        lock: threading.Lock
        num_rows: mpu.ValueProxy[int]

    def __init__(
        self,
        db_url: str,
        ctx: Context,
    ) -> None:
        super().__init__()
        self._db_url = db_url
        self._ctx = ctx

        self._es = contextlib.ExitStack()

        self._num_rows = 0
        self._rows: list[dict] = []

    #

    def __enter__(self) -> ta.Self:
        self._es.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._es.__exit__(exc_type, exc_val, exc_tb)

    #

    @cached.function
    def _db(self) -> sa.Engine:
        db = sa.create_engine(self._db_url)
        self._es.enter_context(lang.defer(db.dispose))  # noqa
        return db

    def _setup_db(self) -> None:
        with self._ctx.lock:
            with self._db().begin() as conn:
                meta.create_all(bind=conn)

    #

    def _flush_rows(self) -> None:
        with self._ctx.lock:
            while True:
                try:
                    with self._db().begin() as conn:
                        conn.execute(pages_table.insert(), self._rows)

                except sa.exc.OperationalError as oe:
                    if 'database is locked' in repr(oe):  # FIXME: lol
                        log.warning(f'database is locked!!')  # noqa
                        time.sleep(1.)
                        continue
                    raise

                else:
                    break

            self._ctx.num_rows.set(self._ctx.num_rows.get() + len(self._rows))
            log.info(
                f'{len(self._rows):_} rows batched, '  # noqa
                f'{self._num_rows:_} rows file, '  # noqa
                f'{self._ctx.num_rows.get():_} rows total'  # noqa
            )

            self._rows.clear()

    row_batch_size = 1_000

    def _maybe_flush_rows(self):
        if len(self._rows) >= self.row_batch_size:
            self._flush_rows()

    #

    verbose = False

    parse_backend: ta.Any = (
        mfh
        # wtp
    )

    def _handle_line(self, l: str) -> bool:
        page = msh.unmarshal(json.loads(l), mdl.Page)  # noqa

        self.verbose and print(page.title)

        if not page.revisions:
            self.verbose and print()
            return False

        rev = page.revisions[0]

        if len(page.revisions) > 1:
            raise Exception(f'Multiple revisions: {page.id=} {page.title=}')

        if not (rev.text and rev.text.text):
            self.verbose and print()
            return False

        page_buf = json.dumps_compact(msh.marshal(page)).encode('utf-8')
        page_compressed_buf = lz4.frame.compress(page_buf)

        row = dict(
            id=page.id,
            title=page.title,
            page=page_compressed_buf,
        )

        if self.parse_backend is mfh:
            doc = mfh.parse_doc(rev.text.text)  # noqa
            # print(doc)

            doc_buf = json.dumps_compact(msh.marshal(doc, mfh.Doc)).encode('utf-8')
            doc_compressed_buf = lz4.frame.compress(doc_buf)

            row.update(
                doc=doc_compressed_buf,
            )

        elif self.parse_backend is wtp:
            tree = wtp.parse_tree(rev.text.text)  # noqa
            # print(tree)

        else:
            raise TypeError(self.parse_backend)

        self._rows.append(row)

        self.verbose and print()
        return True

    @logs.error_logging(log)
    def run(self, file_name: str) -> None:
        log.info(f'{self._ctx.deathpact=} {file_name}')  # noqa

        self._setup_db()

        with contextlib.ExitStack() as es:
            f = es.enter_context(open(file_name, 'rb'))
            # fpr = iou.FileProgressReporter(f, time_interval=5)

            # proc = subprocess.Popen(['lz4', '-cdk', fn], stdout=subprocess.PIPE)
            # f = proc.stdout
            # fpr = None

            zf = es.enter_context(lz4.frame.open(f))
            tw = io.TextIOWrapper(zf, 'utf-8')

            file_pr = ProgressReporter[int](
                fn=f.tell,
                total=os.fstat(f.fileno()).st_size,
                fmt='_',
                suffix='B',
                report_interval=1_000_000,
            )
            rows_pr = ProgressReporter[int](
                start=self._num_rows,
                fmt='_',
                suffix='pages',
            )

            while (l := tw.readline()):
                self._num_rows += 1

                self._ctx.deathpact.poll()

                file_pr.update()
                rows_pr.update(self._num_rows)
                if file_pr.should_report or rows_pr.should_report:
                    log.info(
                        f'{os.path.basename(file_name)}: ' +  # noqa
                        ', '.join([
                            *file_pr.report(),
                            *rows_pr.report(),
                        ]),
                    )

                # if fpr is not None and fpr.report():  # noqa
                #     print(f'{i} pages', file=sys.stderr)

                if self._handle_line(l):
                    self._maybe_flush_rows()

            self._flush_rows()


@logs.error_logging(log)
def analyze_file(
        file_name: str,
        db_url: str,
        ctx: FileAnalyzer.Context,
) -> None:
    FileAnalyzer(db_url, ctx).run(file_name)


##


def _init_process() -> None:
    logs.configure_standard_logging('INFO')


def _main() -> None:
    logs.configure_standard_logging('INFO')

    default_workers = 2

    #

    parser = argparse.ArgumentParser()
    parser.add_argument('-j', '--num-workers', type=int, default=default_workers)
    args = parser.parse_args()

    #

    db_file = os.path.join('.data/wiki.db')
    if os.path.isfile(db_file):
        os.unlink(db_file)
    db_url = f'sqlite:///{db_file}'

    log.info('Launching')  # noqa

    #

    with contextlib.ExitStack() as es:
        if default_workers:
            deathpact: deathpacts.PipeDeathpact = es.enter_context(deathpacts.PipeDeathpact())

            mp_context = mpu.ExtrasSpawnContext(mpu.SpawnExtras(
                pass_fds={deathpact.pass_fd},
                deathsig=signal.SIGTERM,
            ))

            mp_manager = mp_context.Manager()

            ctx = FileAnalyzer.Context(
                deathpact=deathpact,
                lock=mp_manager.Lock(),
                num_rows=mp_manager.Value('i', 0),
            )

            ex = es.enter_context(cfu.new_executor(  # noqa
                args.num_workers,
                cf.ProcessPoolExecutor,
                mp_context=mp_context,
                initializer=_init_process,
            ))

        else:
            ctx = FileAnalyzer.Context(
                deathpact=deathpacts.NopDeathpact(),
                lock=threading.RLock(),  # type: ignore
                num_rows=mpu.DummyValueProxy(0),
            )

            ex = es.enter_context(cfu.ImmediateExecutor())

        futs: list[cf.Future] = [
            ex.submit(
                analyze_file,
                file_name,
                db_url,
                ctx,
            )
            for file_name in sorted(
                glob.glob(os.path.join(LZ4_JSONL_DIR, '*.jsonl.lz4')),
                key=lambda fn: -os.stat(fn).st_size,
            )
        ]

        for fut in cf.as_completed(futs):
            try:
                fut.result()
            except Exception as e:  # noqa
                log.exception('Master caught exception')
                ex.shutdown(wait=False, cancel_futures=True)
                raise

        log.info(f'Complete! {ctx.num_rows.get()} rows total')  # noqa


if __name__ == '__main__':
    _main()



"""
https://github.com/TrueBrain/TrueWiki
https://github.com/TrueBrain/wikitexthtml

  Argument
  Bold
  Comment
  ExternalLink
  Italic
  Parameter
  ParserFunction
  Section
  Table
  Tag
  Template
  WikiLink
  WikiList
  WikiText
"""
import dataclasses as dc
import heapq
import itertools
import typing as ta

from omlish import cached
from omlish import lang


if ta.TYPE_CHECKING:
    import wikitextparser as wtp
else:
    wtp = lang.proxy_import('wikitextparser')


WikiText: ta.TypeAlias = 'wtp.WikiText'


##


@dc.dataclass(frozen=True)
class WtpNode:
    wiki: 'wtp.WikiText'
    children: list[ta.Union['WtpNode', str]] = dc.field(default_factory=list)

    @cached.property
    def span(self) -> tuple[int, int]:
        return self.wiki.span  # noqa


def parse_tree(src: str) -> WtpNode:
    wiki = wtp.parse(src)

    part_its = [
        wiki.parameters,
        wiki.parser_functions,
        wiki.templates,
        wiki.wikilinks,
        wiki.comments,
        wiki.get_bolds_and_italics(),
        wiki.external_links,
        wiki.sections,
        wiki.get_tables(recursive=True),
        wiki.get_lists(),
        wiki.get_tags(),
    ]

    #

    flat_it = itertools.chain.from_iterable(
        sorted(
            [o for _, _, o in g],
            key=lambda o: -o.span[1],
        )
        for _, g in itertools.groupby(
            heapq.merge(*[
                (lambda i, it: ((o.span, i, o) for o in it))(i, it)  # noqa
                for i, it in enumerate(part_its)
            ]),
            key=lambda t: t[0][0],
        )
    )

    #

    root = WtpNode(wiki)
    stk: list[WtpNode] = [root]

    o: wtp.WikiText
    for o in flat_it:
        p = None
        while o.span[0] >= stk[-1].wiki.span[1]:
            p = stk.pop()
            stk[-1].children.append(p)

        if p is not None:
            l, r = p.wiki.span[1], o.span[0]
            if l > r:
                raise Exception(f'{p.wiki.span=} {o.span=}')
            if (r - l) > 1:
                stk[-1].children.append(src[l:r])

        stk.append(WtpNode(o))

    while len(stk) > 1:
        p = stk.pop()
        stk[-1].children.append(p)

    if stk.pop() is not root:
        raise RuntimeError

    return root



"""
https://stackoverflow.com/a/78339316

git clone --single-branch --branch=master --filter=blob:none --sparse --depth=1 https://github.com/ggerganov/llama.cpp
cd llama.cpp
git sparse-checkout set --no-cone '/*' '!/models/*'
git checkout 8f1d81a0b6f50b9bad72db0b6fcd299ad9ecd48c

git submodule status
-8f1d81a0b6f50b9bad72db0b6fcd299ad9ecd48c vendor/llama.cpp
"""
import glob
import os.path
import shutil
import subprocess
import sys
import tempfile


REPO_URL = 'https://github.com/abetlen/llama-cpp-python'

REV: str | None = None
# REV = 'v0.2.90'
# REV = '077ecb6771fbd373d5507444f6e0b6e9bb7cf4e8'

MINIMAL_SUBMODULE = True


def _main() -> None:
    tmp_dir = tempfile.mkdtemp(f'-{__package__.replace(".", "-")}-buildwheel')

    print(tmp_dir)

    #

    dir_name = 'llama-cpp-python'

    subprocess.check_call([
        'git', 'clone', '--depth=1', REPO_URL, dir_name,
    ], cwd=tmp_dir)
    repo_dir = os.path.join(tmp_dir, dir_name)

    if REV is not None:
        subprocess.check_call([
            'git', 'fetch', '--tags', 'origin', REV,
        ], cwd=repo_dir)
        subprocess.check_call([
            'git', 'checkout', REV,
        ], cwd=repo_dir)

    #

    if MINIMAL_SUBMODULE:
        sub_rev = subprocess.check_output([
            'git', 'submodule', 'status',
        ], cwd=repo_dir).decode().strip()[1:].split()[0]
        print(sub_rev)

        sub_url = subprocess.check_output([
            'git',
            'config',
            '--file=.gitmodules',
            'submodule.vendor/llama.cpp.url',
        ], cwd=repo_dir).decode().strip()
        print(sub_url)

        vendor_dir = os.path.join(repo_dir, 'vendor')
        sub_dir = os.path.join(vendor_dir, 'llama.cpp')
        shutil.rmtree(sub_dir)

        subprocess.check_call([
            'git',
            'clone',
            '--single-branch',
            '--branch=master',
            '--filter=blob:none',
            '--sparse',
            '--depth=1',
            sub_url,
            'llama.cpp',
        ], cwd=vendor_dir)

        subprocess.check_call([
            'git',
            'sparse-checkout',
            'set',
            '--no-cone',
            '/*',
            # '!/examples/*',
            '!/models/*',
        ], cwd=sub_dir)

        subprocess.check_call([
            'git', 'checkout', sub_rev,
        ], cwd=sub_dir)

        os.unlink(os.path.join(repo_dir, '.gitmodules'))

    else:
        subprocess.check_call([
            'git', 'submodule', 'update', '--init',
        ], cwd=repo_dir)

    #

    subprocess.check_call([
        sys.executable, '-m', 'venv', '.venv',
    ], cwd=repo_dir)
    venv_exe_file = os.path.join(repo_dir, '.venv', 'bin', 'python')

    subprocess.check_call([
        venv_exe_file, '-m', 'pip', 'install', 'build', 'wheel',
    ], cwd=repo_dir)

    subprocess.check_call([
        venv_exe_file, '-m', 'pip', 'install', '-e', '.[dev]',
    ], cwd=repo_dir)

    #

    path_items: list[str] = []
    cmake_args: list[str] = []

    if sys.platform == 'linux':
        cuda_dir = os.path.join(os.path.realpath('/usr/local/cuda'), 'bin')
        path_items.append(cuda_dir)
        cmake_args.append('-DGGML_CUDA=on')

    elif sys.platform == 'darwin':
        cmake_args.append('-DGGML_METAL=on')

    build_env = {
        **os.environ,
        'PATH': os.pathsep.join([*path_items, os.environ['PATH']]),
        'CMAKE_ARGS': ' '.join(cmake_args),
    }

    subprocess.check_call(
        [venv_exe_file, '-m', 'build', '--wheel'],
        env=build_env,
        cwd=repo_dir,
    )

    #

    dist_dir = os.path.join(repo_dir, 'dist')
    whl_files = glob.glob(os.path.join(dist_dir, '*.whl'))
    print('\n'.join(whl_files))


if __name__ == '__main__':
    _main()



import os
import sys
import typing as ta

from omlish import lang

from ..prompts import PromptModel
from ..prompts import PromptRequest
from ..prompts import PromptResponse


if ta.TYPE_CHECKING:
    import transformers
else:
    transformers = lang.proxy_import('transformers')


class TransformersPromptModel(PromptModel):
    DEFAULT_MODEL: ta.ClassVar[str] = (
        'microsoft/phi-2'
        # 'Qwen/Qwen2-0.5B'
        # 'meta-llama/Meta-Llama-3-8B'
    )

    def __init__(
            self,
            model: str = DEFAULT_MODEL,
            kwargs: ta.Mapping[str, ta.Any] | None = None,
            *,
            token: str | None = None,
    ) -> None:
        super().__init__()
        self._model = model
        self._kwargs = kwargs
        self._token = token

    def invoke(self, request: PromptRequest) -> PromptResponse:
        pipeline = transformers.pipeline(
            'text-generation',
            model=self._model,
            **{
                **dict(
                    device='mps' if sys.platform == 'darwin' else 'cuda',
                    token=self._token or os.environ.get('HUGGINGFACE_HUB_TOKEN'),
                ),
                **(self._kwargs or {}),
            },
        )
        output = pipeline(request.v)
        return PromptResponse(v=output)



from omlish.__about__ import ProjectBase
from omlish.__about__ import SetuptoolsBase
from omlish.__about__ import __version__


class Project(ProjectBase):
    name = 'ommlx'
    description = 'ommlx'

    dependencies = [
        f'omdev == {__version__}',
        f'omlish == {__version__}',
    ]

    optional_dependencies = {
        'backends': [
            'mlx-lm ~= 0.22; sys_platform == "darwin"',

            'sentence-transformers ~= 3.4',

            'transformers ~= 4.49',
        ],

        'huggingface': [
            'huggingface-hub ~= 0.27',
            'datasets ~= 3.1',
        ],

        'llamacpp': [
            'llama-cpp-python ~= 0.3',
        ],

        'numpy': [
            'numpy >= 1.20',
        ],

        'ocr': [
            'pytesseract ~= 0.3',

            'rapidocr-onnxruntime ~= 1.4',
        ],

        'pillow': [
            'pillow ~= 11.0',
        ],

        'search': [
            'duckduckgo-search ~= 7.3',
        ],

        'tinygrad': [
            # 'tinygrad @ git+https://github.com/tinygrad/tinygrad',
            'tinygrad ~= 0.9',
        ],

        'torch': [
            'torch ~= 2.6',
        ],

        'wiki': [
            'mwparserfromhell ~= 0.6',

            'wikitextparser ~= 0.56',  # !! GPL
        ],

        'xml': [
            'lxml ~= 5.3; python_version < "3.13"',
        ],
    }

    entry_points = {
        'omlish.manifests': {name: name},
    }


class Setuptools(SetuptoolsBase):
    find_packages = {
        'include': [Project.name, f'{Project.name}.*'],
        'exclude': [*SetuptoolsBase.find_packages['exclude']],
    }



"""
TODO:
 - defaults:
   {
     "frequency_penalty": 0.0,
     "max_tokens": 64,
     "messages": ...,
     "model": "gpt-4o",
     "presence_penalty": 0.0,
     "temperature": 0.1,
     "top_p": 1
   }
"""
import os
import typing as ta

from omlish import check
from omlish.formats import json
from omlish.http import all as http
from omlish.secrets.secrets import Secret

from ...chat.messages import AiMessage
from ...chat.messages import Message
from ...chat.messages import SystemMessage
from ...chat.messages import ToolExecRequest
from ...chat.messages import ToolExecResultMessage
from ...chat.messages import UserMessage
from ...chat.models import AiChoice
from ...chat.models import ChatModel
from ...chat.models import ChatRequest
from ...chat.models import ChatResponse
from ...chat.options import ChatOptions
from ...chat.tools import Tool
from ...chat.tools import ToolSpec
from ...generative import MaxTokens
from ...generative import Temperature
from ...models import TokenUsage
from ...options import Options
from ...options import ScalarOption


def _opt_dct_fld(k, v):
    return {k: v} if v else {}


def render_tool_spec(ts: ToolSpec) -> ta.Mapping[str, ta.Any]:
    return dict(
        name=ts.name,

        **_opt_dct_fld('description', ts.desc),

        parameters=dict(
            type='object',
            properties={
                tp.name: {
                    'type': tp.dtype,
                    **_opt_dct_fld('description', tp.desc),
                }
                for tp in ts.params
            },

            required=[tp.name for tp in ts.params if tp.required],
            additionalProperties=False,
        ),
    )


def build_request_message(m: Message) -> ta.Mapping[str, ta.Any]:
    if isinstance(m, SystemMessage):
        return dict(
            role='system',
            content=m.s,
        )

    elif isinstance(m, AiMessage):
        return dict(
            role='assistant',
            content=m.s,
            **(dict(tool_calls=[
                dict(
                    id=te.id,
                    function=dict(
                        arguments=te.args,
                        name=te.spec.name,
                    ),
                    type='function',
                )
                for te in m.tool_exec_requests
            ]) if m.tool_exec_requests else {}),
        )

    elif isinstance(m, UserMessage):
        return dict(
            role='user',
            content=check.isinstance(m.c, str),
        )

    elif isinstance(m, ToolExecResultMessage):
        return dict(
            role='tool',
            tool_call_id=m.id,
            content=m.s,
        )

    else:
        raise TypeError(m)


class OpenaiChatModel(ChatModel):
    DEFAULT_MODEL: ta.ClassVar[str] = (
        # 'gpt-4o'
        'gpt-4o-mini'
    )

    ROLES_MAP: ta.ClassVar[ta.Mapping[type[Message], str]] = {
        SystemMessage: 'system',
        UserMessage: 'user',
        AiMessage: 'assistant',
        ToolExecResultMessage: 'tool',
    }

    DEFAULT_OPTIONS: ta.ClassVar[Options[ChatOptions]] = Options(
        Temperature(0.),
        MaxTokens(1024),
    )

    def __init__(
            self,
            *,
            model: str | None = None,
            api_key: Secret | str | None = None,
    ) -> None:
        super().__init__()
        self._model = model or self.DEFAULT_MODEL
        self._api_key = Secret.of(api_key if api_key is not None else os.environ['OPENAI_API_KEY'])

    _OPTION_KWARG_NAMES_MAP: ta.Mapping[type[ScalarOption], str] = {
        Temperature: 'temperature',
        MaxTokens: 'max_tokens',
    }

    def invoke(self, request: ChatRequest) -> ChatResponse:
        kw: dict = dict(
            temperature=0,
            max_tokens=1024,
        )

        tools_by_name: dict[str, ToolSpec] = {}

        for opt in request.options:
            if isinstance(opt, ScalarOption) and (kwn := self._OPTION_KWARG_NAMES_MAP.get(type(opt))) is not None:
                kw[kwn] = opt.v

            elif isinstance(opt, Tool):
                if opt.spec.name in tools_by_name:
                    raise NameError(opt.spec.name)
                tools_by_name[opt.spec.name] = opt.spec

            else:
                raise TypeError(opt)

        tools = [
            dict(
                type='function',
                function=render_tool_spec(ts),
            )
            for ts in tools_by_name.values()
        ]

        raw_request = dict(
            model=self._model,
            messages=[
                build_request_message(m)
                for m in request.v
            ],
            top_p=1,
            **(dict(tools=tools) if tools else {}),
            frequency_penalty=0.0,
            presence_penalty=0.0,
            stream=False,
            **kw,
        )

        raw_response = http.request(
            'https://api.openai.com/v1/chat/completions',
            headers={
                http.consts.HEADER_CONTENT_TYPE: http.consts.CONTENT_TYPE_JSON,
                http.consts.HEADER_AUTH: http.consts.format_bearer_auth_header(check.not_none(self._api_key).reveal()),
            },
            data=json.dumps(raw_request).encode('utf-8'),
        )

        response = json.loads(check.not_none(raw_response.data).decode('utf-8'))

        return ChatResponse(
            v=[
                AiChoice(AiMessage(
                    choice['message']['content'],
                    tool_exec_requests=[
                        ToolExecRequest(
                            id=tc['id'],
                            spec=tools_by_name[tc['function']['name']],
                            args=tc['function']['arguments'],
                        )
                        for tc in choice['message'].get('tool_calls', [])
                    ],
                ))
                for choice in response['choices']
            ],
            usage=TokenUsage(
                input=response['usage']['prompt_tokens'],
                output=response['usage']['completion_tokens'],
                total=response['usage']['total_tokens'],
            ) if response.get('usage') is not None else None,
        )



import contextlib
import logging
import os.path
import signal
import typing as ta

from omlish import check
from omlish.argparse import all as ap
from omlish.daemons.spawning import ForkSpawning  # noqa
from omlish.daemons.spawning import MultiprocessingSpawning  # noqa
from omlish.daemons.spawning import ThreadSpawning  # noqa
from omlish.logs import all as logs
from omlish.os.pidfiles.pinning import PidfilePinner
from omlish.os.signals import parse_signal

from .client import McServerClient
from .service import mc_server_service_daemon


log = logging.getLogger(__name__)


##


class McServerCli(ap.Cli):
    @ap.cmd()
    def launch(self) -> None:
        mc_server_service_daemon().daemon_().launch()

    @ap.cmd(
        ap.arg('prompt'),
        ap.arg('--no-launch', action='store_true'),
    )
    def prompt(self) -> None:
        print(McServerClient().prompt(
            self.args.prompt,
            launch=not self.args.no_launch,
        ))

    #

    @contextlib.contextmanager
    def _pin_pid(self) -> ta.Iterator[int]:
        pid_file = check.non_empty_str(mc_server_service_daemon().daemon_config().pid_file)
        with PidfilePinner.default_impl()().pin_pidfile_owner(pid_file) as pid:
            yield pid

    @ap.cmd()
    def pid(self) -> None:
        with self._pin_pid() as pid:
            print(pid)

    @ap.cmd(
        ap.arg('signal', nargs='?'),
    )
    def kill(self) -> None:
        if (sig_arg := self.args.signal) is not None:
            sig = parse_signal(sig_arg)
        else:
            sig = signal.SIGTERM

        with self._pin_pid() as pid:
            log.info('Killing pid: %d', pid)
            os.kill(pid, sig)


##


def _main() -> None:
    logs.configure_standard_logging('DEBUG')

    McServerCli().cli_run_and_exit()


if __name__ == '__main__':
    _main()



"""
TODO:
 - use omserv
  - core daemon service skeleton should interop with simple server
"""
import dataclasses as dc
import logging
import time
import typing as ta

from omdev.home.secrets import load_secrets
from omlish import cached
from omlish import check
from omlish.http.coro.simple import make_simple_http_server
from omlish.http.handlers import ExceptionLoggingHttpHandler
from omlish.http.handlers import HttpHandler_
from omlish.http.handlers import HttpHandlerRequest
from omlish.http.handlers import HttpHandlerResponse
from omlish.http.handlers import LoggingHttpHandler
from omlish.sockets.bind import CanSocketBinderConfig
from omlish.sockets.bind import SocketBinder
from omlish.sockets.server.server import SocketServer

from .. import minichain as mc
from ..minichain.backends.mlxlm import MlxlmChatModel
from ..minichain.backends.openai import OpenaiChatModel
from ..minichain.generative import Temperature


log = logging.getLogger(__name__)


##


@dc.dataclass(frozen=True)
class McServerHandler(HttpHandler_):
    llm: mc.ChatModel

    def __call__(self, req: HttpHandlerRequest) -> HttpHandlerResponse:
        prompt = check.not_none(req.data).decode('utf-8')

        log.info('Server got prompt: %s', prompt)

        resp = self.llm(
            [mc.UserMessage(prompt)],
            Temperature(.1),
        )
        resp_txt = check.not_none(resp.v[0].m.s)

        log.info('Server got response: %s', resp_txt)

        data = resp_txt.encode('utf-8')
        return HttpHandlerResponse(
            status=200,
            headers={
                'Content-Length': str(len(data)),
            },
            data=data,
            close_connection=True,
        )


class McServer:
    @dc.dataclass(frozen=True)
    class Config:
        DEFAULT_BIND: ta.ClassVar[CanSocketBinderConfig] = 5067
        bind: SocketBinder.Config = SocketBinder.Config.of(DEFAULT_BIND)

        backend: ta.Literal['openai', 'local'] = 'local'

        linger_s: float | None = 10 * 60.

    def __init__(self, config: Config = Config()) -> None:
        super().__init__()

        self._config = config

    @cached.function
    def llm(self) -> mc.ChatModel:
        if self._config.backend == 'openai':
            return OpenaiChatModel(api_key=load_secrets().get('openai_api_key').reveal())

        elif self._config.backend == 'local':
            model = 'mlx-community/Qwen2.5-Coder-32B-Instruct-8bit'
            return MlxlmChatModel(model)

        else:
            raise ValueError(self._config.backend)

    def run(self) -> None:
        log.info('Server running')

        try:
            llm = self.llm()

            with make_simple_http_server(
                    self._config.bind,
                    ExceptionLoggingHttpHandler(
                        LoggingHttpHandler(
                            McServerHandler(llm),
                            log,
                        ),
                        log,
                    ),
            ) as server:
                if (linger_s := self._config.linger_s) is None:
                    linger_s = float('inf')

                deadline = time.monotonic() + linger_s

                with server.poll_context() as pc:
                    while True:
                        if time.monotonic() >= deadline:
                            log.info('Linger deadline exceeded')
                            break

                        res = pc.poll(5.)

                        if res in (SocketServer.PollResult.ERROR, SocketServer.PollResult.SHUTDOWN):
                            break

                        if res == SocketServer.PollResult.CONNECTION:
                            deadline = time.monotonic() + linger_s

        finally:
            log.info('Server exiting')

    @classmethod
    def run_config(cls, config: Config) -> None:
        return cls(config).run()



"""
https://docs.mistral.ai/getting-started/models/
"""
import os
import typing as ta

from omlish import check
from omlish.formats import json
from omlish.http import all as http

from ..chat.messages import AiMessage
from ..chat.messages import Message
from ..chat.messages import SystemMessage
from ..chat.messages import UserMessage
from ..chat.models import AiChoice
from ..chat.models import ChatModel
from ..chat.models import ChatRequest
from ..chat.models import ChatResponse


class MistralChatModel(ChatModel):
    model: ta.ClassVar[str] = 'mistral-large-latest'

    ROLES_MAP: ta.ClassVar[ta.Mapping[type[Message], str]] = {
        SystemMessage: 'system',
        UserMessage: 'user',
        AiMessage: 'assistant',
    }

    def __init__(self, *, api_key: str | None = None) -> None:
        super().__init__()
        self._api_key = api_key

    def _get_msg_content(self, m: Message) -> str | None:
        if isinstance(m, (SystemMessage, AiMessage)):
            return m.s

        elif isinstance(m, UserMessage):
            return check.isinstance(m.c, str)

        else:
            raise TypeError(m)

    def invoke(
            self,
            request: ChatRequest,
    ) -> ChatResponse:
        if not (key := self._api_key):
            key = os.environ['MISTRAL_API_KEY']

        req_dct = {
            'model': self.model,
            'messages': [
                {
                    'role': self.ROLES_MAP[type(m)],
                    'content': self._get_msg_content(m),
                }
                for m in request.v
            ],
        }

        resp = http.request(
            'https://api.mistral.ai/v1/chat/completions',
            method='POST',
            data=json.dumps_compact(req_dct).encode('utf-8'),
            headers={
                'Content-Type': 'application/json',
                'Accept': 'application/json',
                'Authorization': f'Bearer {key}',
            },
        )

        resp_dct = json.loads(check.not_none(resp.data).decode('utf-8'))

        return ChatResponse(v=[
            AiChoice(AiMessage(c['message']['content']))
            for c in resp_dct['choices']
        ])



import typing as ta

from omlish import check

from . import models as mdl
from .utils import xml as xu


T = ta.TypeVar('T')


##


def symm_dct(*ks: T) -> ta.Mapping[T, T]:
    return {k: k for k in ks}


def parse_int(s: str | None) -> int:
    return int(check.non_empty_str(s))


##


parse_namespace = xu.ElementToObj(
    mdl.Namespace,
    xu.ElementToKwargs(
        attrs={
            'key': ('key', int),
            'case': 'case',
        },
        text='text',
    ),
)


parse_namespaces = xu.ElementToObj(
    mdl.Namespaces,
    xu.ElementToKwargs(
        list_children={
            'namespace': ('namespaces', parse_namespace),
        },
    ),
)


parse_site_info = xu.ElementToObj(
    mdl.SiteInfo,
    xu.ElementToKwargs(
        scalars=symm_dct(
            'sitename',
            'dbname',
            'base',
            'generator',
            'case',
        ),
        single_children={
            'namespaces': ('namespaces', parse_namespaces),
        },
    ),
)


parse_redirect = xu.ElementToObj(
    mdl.Redirect,
    xu.ElementToKwargs(
        attrs=symm_dct(
            'title',
        ),
    ),
)


def _parse_deleted(s: str) -> bool:
    if s != 'deleted':
        raise ValueError(s)
    return True


parse_contributor = xu.ElementToObj(
    mdl.Contributor,
    xu.ElementToKwargs(
        attrs={
            'deleted': ('deleted', _parse_deleted),
        },
        scalars={
            **symm_dct(
                'username',
                'ip',
            ),
            'id': ('id', parse_int),
        },
    ),
)


parse_revision_text = xu.ElementToObj(
    mdl.RevisionText,
    xu.ElementToKwargs(
        attrs={
            'bytes': ('bytes', int),
            'sha1': 'sha1',
            'space': None,
        },
        text='text',
    ),
)


parse_revision = xu.ElementToObj(
    mdl.Revision,
    xu.ElementToKwargs(
        scalars={
            'id': ('id', parse_int),
            'parentid': ('parentid', parse_int),
            'origin': ('origin', parse_int),
            'minor': ('minor', lambda _: True),
            **symm_dct(
                'timestamp',
                'comment',
                'model',
                'format',
                'sha1',
            ),
        },
        single_children={
            'text': ('text', parse_revision_text),
        },
        list_children={
            'contributor': ('contributors', parse_contributor),
        },
    ),
)


parse_page = xu.ElementToObj(
    mdl.Page,
    xu.ElementToKwargs(
        scalars={
            'ns': ('ns', parse_int),
            'id': ('id', parse_int),
            **symm_dct(
                'title',
                'restrictions',
            ),
        },
        single_children={
            'redirect': ('redirect', parse_redirect),
        },
        list_children={
            'revision': ('revisions', parse_revision),
        },
    ),
)



import collections
import os.path
import typing as ta

from omdev.cache import data as dcache
from omlish import cached
from omlish import dataclasses as dc
from omlish.formats import json


@dc.dataclass(frozen=True)
class Movie:
    name: str
    dct: ta.Mapping[str, ta.Any]
    links: ta.Sequence[str]
    rat_pct: str
    rat_10: str


MOVIES_DATA = dcache.GithubContentSpec(
    'DOsinga/deep_learning_cookbook',
    '04f56a7fe11e16c19ec6269bc5a138efdcb522a7',
    ['data/wp_movies_10k.ndjson'],
)


def load_movies(cache_dir: str) -> ta.Sequence[Movie]:
    movies = []

    if not os.path.exists(cache_dir):
        os.mkdir(cache_dir)

    data_file = dcache.default().get(MOVIES_DATA)

    with open(data_file) as f:
        for l in f:
            movies.append(Movie(*json.loads(l)))

    return movies


@dc.dataclass(frozen=True)
class MoviesData:
    movies: ta.Sequence[Movie]

    @cached.property
    def top_links(self) -> ta.Sequence[str]:
        link_counts = collections.Counter[str]()
        for movie in self.movies:
            link_counts.update(movie.links)
        return [link for link, c in link_counts.items() if c >= 3]

    @cached.property
    def link_to_idx(self) -> ta.Mapping[str, int]:
        return {link: idx for idx, link in enumerate(self.top_links)}

    @cached.property
    def movie_to_idx(self) -> ta.Mapping[str, int]:
        return {movie.name: idx for idx, movie in enumerate(self.movies)}

    @cached.property
    def pairs(self) -> ta.Sequence[tuple[int, int]]:
        pairs: list[tuple[int, int]] = []
        for movie in self.movies:
            pairs.extend(
                (self.link_to_idx[link], self.movie_to_idx[movie.name])
                for link in movie.links
                if link in self.link_to_idx
            )
        return pairs

    @cached.property
    def pairs_set(self) -> ta.AbstractSet[tuple[int, int]]:
        pairs_set = set(self.pairs)
        return pairs_set



import collections.abc
import io
import typing as ta

from omlish import dispatch

from .content import Content
from .images import Image
from .placeholders import Env
from .placeholders import Placeholder


##


class PlaceholderFiller:
    def __init__(self, env: Env) -> None:
        super().__init__()
        self._env = env

    @dispatch.method
    def apply(self, c: Content) -> Content:
        raise TypeError(c)

    @apply.register
    def apply_str(self, c: str) -> str:
        return c

    @apply.register
    def apply_sequence(self, s: collections.abc.Sequence) -> collections.abc.Sequence:
        return [self.apply(e) for e in s]

    @apply.register
    def apply_image(self, c: Image) -> Image:
        return c

    @apply.register
    def apply_placeholder(self, c: Placeholder) -> str:
        return self._env[c.k]


##


class StringRenderer:
    def __init__(self, out: ta.TextIO) -> None:
        super().__init__()

        self._out = out

    @dispatch.method
    def render(self, c: Content) -> None:
        raise TypeError(c)

    @render.register
    def render_str(self, c: str) -> None:
        self._out.write(c)

    @render.register
    def render_sequence(self, s: collections.abc.Sequence) -> None:
        for e in s:
            self.render(e)

    @classmethod
    def render_to_str(cls, c: Content) -> str:
        out = io.StringIO()
        cls(out).render(c)
        return out.getvalue()



"""
TODO:
  - lazy load
  - serialize fs path not data
"""
import typing as ta

from omlish import dataclasses as dc
from omlish import lang

from .content import ExtendedContent


if ta.TYPE_CHECKING:
    import PIL.Image as pimg  # noqa
else:
    pimg = lang.proxy_import('PIL.Image')


##


@dc.dataclass(frozen=True)
class Image(ExtendedContent, lang.Final):
    i: 'pimg.Image' = dc.field()



from omlish import lang

from .options import Option
from .options import ScalarOption
from .options import UniqueOption


##


class Generative(lang.Abstract):
    pass


##


class GenerativeOption(Option, lang.Abstract):
    pass


class TopK(GenerativeOption, UniqueOption, ScalarOption[int], lang.Final):
    pass


class Temperature(GenerativeOption, UniqueOption, ScalarOption[float], lang.Final):
    pass


class MaxTokens(GenerativeOption, UniqueOption, ScalarOption[int], lang.Final):
    pass



import typing as ta

from omlish import cached
from omlish import collections as col
from omlish import dataclasses as dc
from omlish import lang

from .options import ChatOption


ToolDtype: ta.TypeAlias = str


##


@dc.dataclass(frozen=True)
class ToolParam(lang.Final):
    name: str
    dtype: ToolDtype

    _: dc.KW_ONLY

    desc: str | None = None
    required: bool = False


@dc.dataclass(frozen=True)
class ToolSpec(lang.Final):
    name: str
    params: ta.Sequence[ToolParam]

    _: dc.KW_ONLY

    desc: str

    @cached.property
    @dc.init
    def params_by_name(self) -> ta.Mapping[str, ToolParam]:
        return col.make_map_by(lambda p: p.name, self.params, strict=True)


@dc.dataclass(frozen=True)
class Tool(ChatOption, lang.Final):
    spec: ToolSpec


@dc.dataclass(frozen=True)
class ToolExecRequest(lang.Final):
    id: str
    spec: ToolSpec
    args: str



"""
TODO:
 - marshal
 - __getitem__ return Vector? views? this isn't a replacement for strided tensors, just List[float]
 - memoryview?
 - torch.Tensor?
"""
import array
import functools
import struct
import sys
import typing as ta

from omlish import check
from omlish import dataclasses as dc
from omlish import lang


if ta.TYPE_CHECKING:
    import numpy as np
else:
    np = lang.proxy_import('numpy')


##


VectorStorage: ta.TypeAlias = ta.Union[
    array.array,
    'np.ndarray',
]


Vectorable: ta.TypeAlias = ta.Union[
    'Vector',
    'VectorStorage',
    bytes,
    ta.Iterable[float],
]


##


@dc.dataclass(frozen=True)
class _StorageImpl:
    cls: type[VectorStorage]
    ctor: ta.Callable[[ta.Iterable[float]], VectorStorage]

    _: dc.KW_ONLY

    is_np: bool = False


_ARRAY_STORAGE_IMPL = _StorageImpl(
    array.array,
    functools.partial(array.array, 'f'),
)

_STORAGE_IMPL: _StorageImpl = _ARRAY_STORAGE_IMPL


def _get_storage_impl() -> _StorageImpl:
    if 'numpy' in sys.modules:
        global _STORAGE_IMPL

        if _STORAGE_IMPL is _ARRAY_STORAGE_IMPL:
            _STORAGE_IMPL = _StorageImpl(
                np.ndarray,
                functools.partial(np.array, dtype=np.float32),
                is_np=True,
            )

    return _STORAGE_IMPL  # noqa


##


_STRUCT_FLOATS_FMTS: dict[int, str] = {}


def _get_struct_float_fmt(d: int) -> str:
    try:
        return _STRUCT_FLOATS_FMTS[d]
    except KeyError:
        fmt = _STRUCT_FLOATS_FMTS[d] = '<' + 'f' * d
        return fmt


def _encode_float_bytes(fs: ta.Sequence[float]) -> bytes:
    return struct.pack(_get_struct_float_fmt(len(fs)), *fs)


def _decode_float_bytes(b: bytes) -> ta.Sequence[float]:
    return struct.unpack(_get_struct_float_fmt(len(b) // 4), b)


##


class Vector(lang.Final, ta.Sequence[float]):
    def __init__(self, obj: Vectorable) -> None:
        if isinstance(obj, Vector):
            check.is_(self, obj)
            return

        super().__init__()

        s: VectorStorage
        si = _get_storage_impl()
        if isinstance(obj, si.cls):
            s = obj  # type: ignore
        else:
            l: ta.Iterable[float]
            if isinstance(obj, bytes):
                l = _decode_float_bytes(obj)
            else:
                l = obj
            s = si.ctor(l)

        self._s = s

    def __new__(cls, obj):
        if isinstance(obj, Vector):
            return obj

        return super().__new__(cls)

    #

    def __iter__(self) -> ta.Iterator[float]:
        return iter(self._s)

    @ta.overload
    def __getitem__(self, index: int) -> float:
        ...

    @ta.overload
    def __getitem__(self, index: slice) -> ta.Sequence[float]:
        ...

    def __getitem__(self, index):
        return self._s[index]

    def __len__(self) -> int:
        return len(self._s)

    #

    def bytes(self) -> bytes:
        return _encode_float_bytes(self._s)  # type: ignore

    def np(self) -> 'np.ndarray':
        si = _get_storage_impl()

        if not si.is_np:
            import numpy  # noqa

            si = _get_storage_impl()
            check.state(si.is_np)

        if not isinstance(self._s, si.cls):
            self._s = si.ctor(self._s)

        return self._s  # type: ignore



"""
See:
 - https://github.com/simonw/llm
 - https://github.com/TheR1D/shell_gpt
 - https://github.com/paul-gauthier/aider
"""
import argparse
import dataclasses as dc
import datetime
import os.path
import sys
import typing as ta

from omdev.home.paths import get_home_paths
from omdev.home.secrets import load_secrets
from omlish import check
from omlish import lang
from omlish.diag import pycharm
from omlish.formats import json
from omlish.logs import all as logs

from .. import minichain as mc
from ..minichain.backends.anthropic import AnthropicChatModel
from ..minichain.backends.google import GoogleChatModel
from ..minichain.backends.llamacpp import LlamacppChatModel
from ..minichain.backends.llamacpp import LlamacppPromptModel
from ..minichain.backends.mistral import MistralChatModel
from ..minichain.backends.openai import OpenaiChatModel
from ..minichain.backends.openai import OpenaiEmbeddingModel
from ..minichain.backends.openai import OpenaiPromptModel
from ..minichain.backends.sentencetransformers import SentencetransformersEmbeddingModel
from ..minichain.backends.transformers import TransformersPromptModel
from .state import load_state
from .state import save_state


if ta.TYPE_CHECKING:
    import PIL.Image as pimg  # noqa
else:
    pimg = lang.proxy_import('PIL.Image')


##


DEFAULT_BACKEND = 'openai'


##


@dc.dataclass(frozen=True)
class ChatState:
    name: str | None = None

    created_at: datetime.datetime = dc.field(default_factory=lang.utcnow)
    updated_at: datetime.datetime = dc.field(default_factory=lang.utcnow)

    chat: mc.Chat = ()


CHAT_MODEL_BACKENDS: ta.Mapping[str, type[mc.ChatModel]] = {
    'anthropic': AnthropicChatModel,
    'google': GoogleChatModel,
    'mistral': MistralChatModel,
    'openai': OpenaiChatModel,
    'llamacpp': LlamacppChatModel,
}


def _run_chat(
        content: mc.Content,
        *,
        new: bool = False,
        backend: str | None = None,
) -> None:
    prompt = check.isinstance(content, str)

    state_dir = os.path.join(get_home_paths().state_dir, 'minichain', 'cli')
    if not os.path.exists(state_dir):
        os.makedirs(state_dir, exist_ok=True)
        os.chmod(state_dir, 0o770)  # noqa

    chat_file = os.path.join(state_dir, 'chat.json')
    if new:
        state = ChatState()
    else:
        state = load_state(chat_file, ChatState)  # type: ignore
        if state is None:
            state = ChatState()  # type: ignore

    state = dc.replace(
        state,
        chat=[
            *state.chat,
            mc.UserMessage(prompt),
        ],
    )

    mdl = CHAT_MODEL_BACKENDS[backend or DEFAULT_BACKEND]()
    response = mdl.invoke(mc.ChatRequest.new(state.chat))
    print(check.isinstance(response.v[0].m.s, str).strip())

    chat = dc.replace(
        state,
        chat=[
            *state.chat,
            response.v[0].m,
        ],
    )

    chat = dc.replace(
        chat,
        updated_at=lang.utcnow(),
    )

    save_state(chat_file, chat, ChatState)


##


PROMPT_MODEL_BACKENDS: ta.Mapping[str, type[mc.PromptModel]] = {
    'llamacpp': LlamacppPromptModel,
    'openai': OpenaiPromptModel,
    'transformers': TransformersPromptModel,
}


def _run_prompt(
        content: mc.Content,
        *,
        backend: str | None = None,
) -> None:
    prompt = check.isinstance(content, str)
    mdl = PROMPT_MODEL_BACKENDS[backend or DEFAULT_BACKEND]()
    response = mdl.invoke(mc.PromptRequest.new(prompt))
    print(response.v.strip())


##


EMBEDDING_MODEL_BACKENDS: ta.Mapping[str, type[mc.EmbeddingModel]] = {
    'openai': OpenaiEmbeddingModel,
    'sentencetransformers': SentencetransformersEmbeddingModel,
}


def _run_embed(
        content: mc.Content,
        *,
        backend: str | None = None,
) -> None:
    mdl = EMBEDDING_MODEL_BACKENDS[backend or DEFAULT_BACKEND]()
    response = mdl.invoke(mc.EmbeddingRequest.new(content))
    print(json.dumps_compact(response.v))


##

def _main() -> None:
    logs.configure_standard_logging('INFO')

    #

    parser = argparse.ArgumentParser()
    parser.add_argument('prompt', nargs='+')

    parser.add_argument('-b', '--backend', default='openai')

    parser.add_argument('-c', '--chat', action='store_true')
    parser.add_argument('-n', '--new', action='store_true')
    parser.add_argument('-i', '--interactive', action='store_true')

    parser.add_argument('-e', '--embed', action='store_true')
    parser.add_argument('-j', '--image', action='store_true')

    args = parser.parse_args()

    #

    content: mc.Content

    if args.image:
        content = mc.Image(pimg.open(check.non_empty_str(check.single(args.prompt))))

    else:
        prompt = ' '.join(args.prompt)

        if not sys.stdin.isatty() and not pycharm.is_pycharm_hosted():
            stdin_data = sys.stdin.read()
            prompt = '\n'.join([prompt, stdin_data])

        content = prompt

    #

    # FIXME: lol garbage
    for key in [
        'OPENAI_API_KEY',
        'HUGGINGFACE_TOKEN',
        'TAVILY_API_KEY',
        'ANTHROPIC_API_KEY',
        'MISTRAL_API_KEY',
        'GEMINI_API_KEY',
    ]:
        if (sec := load_secrets().try_get(key.lower())) is not None:
            os.environ[key] = sec.reveal()

    #

    if args.chat:
        _run_chat(
            content,
            backend=args.backend,
            new=bool(args.new),
        )

    elif args.embed:
        _run_embed(
            content,
            backend=args.backend,
        )

    else:
        _run_prompt(
            content,
            backend=args.backend,
        )


if __name__ == '__main__':
    _main()



import urllib.request

from omlish import check
from omlish.daemons.spawning import ForkSpawning  # noqa
from omlish.daemons.spawning import MultiprocessingSpawning  # noqa
from omlish.daemons.spawning import ThreadSpawning  # noqa
from omlish.sockets.bind import TcpSocketBinder
from omlish.sockets.wait import socket_can_connect
from omlish.sockets.wait import socket_wait_until_can_connect

from .service import mc_server_service_daemon


class McServerClient:
    def prompt(
            self,
            prompt: str,
            *,
            launch: bool = False,
            timeout: float = 60.,
    ) -> str:
        ssd = mc_server_service_daemon()

        server_config = ssd.service_config().server
        port = check.isinstance(server_config.bind, TcpSocketBinder.Config).port

        #

        if launch:
            if not socket_can_connect(
                ('localhost', port),
                timeout=timeout,
            ):
                daemon = ssd.daemon_()
                daemon.launch()

                if daemon.config.pid_file is not None:
                    check.state(daemon.is_pidfile_locked())

                socket_wait_until_can_connect(
                    ('localhost', port),
                    timeout=timeout,
                )

        elif not socket_can_connect(
                ('localhost', port),
                timeout=timeout,
        ):
            raise Exception(f'Cannot connect to server on port {port}')

        #

        with urllib.request.urlopen(urllib.request.Request(
                f'http://localhost:{port}/',
                data=prompt.encode('utf-8'),
        )) as resp:
            resp_str = resp.read().decode('utf-8')

        return resp_str



import abc
import typing as ta

from omlish import dataclasses as dc
from omlish import lang

from ..content.content import Content
from ..models import Model
from ..models import ModelOption
from ..models import ModelRequest
from ..models import ModelResponse
from .vectors import Vector


##


EmbeddingInput: ta.TypeAlias = Content
EmbeddingNew: ta.TypeAlias = ta.Any
EmbeddingOutput: ta.TypeAlias = Vector

EmbeddingOptions: ta.TypeAlias = ModelOption


@dc.dataclass(frozen=True, kw_only=True)
class EmbeddingRequest(
    ModelRequest[
        EmbeddingInput,
        EmbeddingOptions,
        EmbeddingNew,
    ],
    lang.Final,
):
    pass


@dc.dataclass(frozen=True, kw_only=True)
class EmbeddingResponse(ModelResponse[EmbeddingOutput], lang.Final):
    pass


class EmbeddingModel(
    Model[
        EmbeddingRequest,
        EmbeddingOptions,
        EmbeddingNew,
        EmbeddingResponse,
    ],
    lang.Abstract,
):
    @abc.abstractmethod
    def invoke(self, request: EmbeddingRequest) -> EmbeddingResponse:
        raise NotImplementedError



import abc
import typing as ta

from omlish import dataclasses as dc
from omlish import lang

from .generative import Generative
from .generative import GenerativeOption
from .models import Model
from .models import ModelOption
from .models import ModelRequest
from .models import ModelResponse
from .services import ServiceOption


##


PromptInput: ta.TypeAlias = str
PromptNew: ta.TypeAlias = str
PromptOutput: ta.TypeAlias = str

PromptOptions: ta.TypeAlias = ServiceOption | ModelOption | GenerativeOption


@dc.dataclass(frozen=True, kw_only=True)
class PromptRequest(
    ModelRequest[
        PromptInput,
        PromptOptions,
        PromptNew,
    ],
    lang.Final,
):
    @dc.validate
    def _validate_v(self) -> bool:
        return isinstance(self.v, str)


@dc.dataclass(frozen=True, kw_only=True)
class PromptResponse(ModelResponse[PromptOutput], lang.Final):
    pass


class PromptModel(
    Model[
        PromptRequest,
        PromptOptions,
        PromptNew,
        PromptResponse,
    ],
    Generative,
    lang.Abstract,
):
    @abc.abstractmethod
    def invoke(self, request: PromptRequest) -> PromptResponse:
        raise NotImplementedError



import contextlib
import os.path
import typing as ta

from omlish import check
from omlish import lang

from ..chat.messages import AiMessage
from ..chat.messages import Message
from ..chat.messages import SystemMessage
from ..chat.messages import ToolExecResultMessage
from ..chat.messages import UserMessage
from ..chat.models import AiChoice
from ..chat.models import ChatModel
from ..chat.models import ChatRequest
from ..chat.models import ChatResponse
from ..prompts import PromptModel
from ..prompts import PromptRequest
from ..prompts import PromptResponse


if ta.TYPE_CHECKING:
    import llama_cpp

    from ... import llamacpp as lcu

else:
    llama_cpp = lang.proxy_import('llama_cpp')

    lcu = lang.proxy_import('...llamacpp', __package__)


class LlamacppPromptModel(PromptModel):
    # hf.hf_hub_download(
    #   revision='1ca85c857dce892b673b988ad0aa83f2cb1bbd19',
    #   repo_id='QuantFactory/Meta-Llama-3-8B-GGUF',
    #   filename='Meta-Llama-3-8B.Q8_0.gguf',
    # )
    model_path = os.path.join(
        os.path.expanduser('~/.cache/huggingface/hub'),
        'models--QuantFactory--Meta-Llama-3-8B-GGUF',
        'snapshots',
        '1ca85c857dce892b673b988ad0aa83f2cb1bbd19',
        'Meta-Llama-3-8B.Q8_0.gguf',
    )

    def invoke(self, request: PromptRequest) -> PromptResponse:
        lcu.install_logging_hook()

        llm = llama_cpp.Llama(
            model_path=self.model_path,
        )

        output = llm.create_completion(
            request.v,
            max_tokens=1024,
            stop=['\n'],
        )

        return PromptResponse(v=output['choices'][0]['text'])  # type: ignore


class LlamacppChatModel(ChatModel):
    model_path = os.path.join(
        os.path.expanduser('~/.cache/huggingface/hub'),
        'models--meta-llama--Llama-3.2-3B-Instruct/snapshots/0cb88a4f764b7a12671c53f0838cd831a0843b95/llama-2-7b-chat.Q5_0.gguf',  # noqa
        # 'models--QuantFactory--Meta-Llama-3-8B-GGUF/snapshots/1ca85c857dce892b673b988ad0aa83f2cb1bbd19/Meta-Llama-3-8B.Q8_0.gguf',  # noqa
    )

    ROLES_MAP: ta.ClassVar[ta.Mapping[type[Message], str]] = {
        SystemMessage: 'system',
        UserMessage: 'user',
        AiMessage: 'assistant',
        ToolExecResultMessage: 'tool',
    }

    def _get_msg_content(self, m: Message) -> str | None:
        if isinstance(m, (SystemMessage, AiMessage)):
            return m.s

        elif isinstance(m, UserMessage):
            return check.isinstance(m.c, str)

        else:
            raise TypeError(m)

    def invoke(self, request: ChatRequest) -> ChatResponse:
        lcu.install_logging_hook()

        with contextlib.ExitStack() as es:
            llm = es.enter_context(contextlib.closing(llama_cpp.Llama(
                model_path=self.model_path,
                verbose=False,
            )))

            output = llm.create_chat_completion(
                messages=[  # noqa
                    dict(  # type: ignore
                        role=self.ROLES_MAP[type(m)],
                        content=self._get_msg_content(m),
                    )
                    for m in request.v
                ],
                max_tokens=1024,
                # stop=['\n'],
            )

            return ChatResponse(v=[
                AiChoice(AiMessage(c['message']['content']))  # noqa
                for c in output['choices']  # type: ignore
            ])



import abc
import typing as ta

from omlish import check
from omlish import dataclasses as dc
from omlish import lang
from omlish import reflect as rfl

from .options import Option
from .options import Options


T = ta.TypeVar('T')
U = ta.TypeVar('U')
ServiceRequestT = ta.TypeVar('ServiceRequestT', bound='ServiceRequest')
ServiceOptionT = ta.TypeVar('ServiceOptionT', bound='Option')
ServiceNewT = ta.TypeVar('ServiceNewT')
ServiceResponseT = ta.TypeVar('ServiceResponseT', bound='ServiceResponse')


##


class ServiceOption(Option, lang.Abstract):
    pass


##


@dc.dataclass(frozen=True, kw_only=True)
class ServiceRequest(lang.Abstract, ta.Generic[T, ServiceOptionT, ServiceNewT]):
    v: T

    options: Options[ServiceOptionT] = dc.xfield(Options(), repr_fn=dc.truthy_repr)

    @classmethod
    def new(
            cls: type[U],
            v: ServiceNewT,
            *options: ServiceOptionT,
            **kwargs: ta.Any,
    ) -> U:
        return cls(
            v=v,  # type: ignore
            options=Options(*options),
            **kwargs,
        )


@dc.dataclass(frozen=True, kw_only=True)
class ServiceResponse(lang.Abstract, ta.Generic[T]):
    v: T


class Service(lang.Abstract, ta.Generic[ServiceRequestT, ServiceOptionT, ServiceNewT, ServiceResponseT]):
    # Not ClassVar - wrappers for example vary by instance.
    request_cls: type[ServiceRequest]
    option_cls_set: frozenset[type[Option]]
    new_cls: ta.Any
    response_cls: type[ServiceResponse]

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        if lang.is_abstract_class(cls):
            return

        gmro = rfl.ALIAS_UPDATING_GENERIC_SUBSTITUTION.generic_mro(cls)

        service_bases = [
            rb
            for rb in gmro
            if isinstance(rb, rfl.Generic)
            and rb.cls is Service
        ]
        if not service_bases:
            return

        service_base = check.single(service_bases)

        present_attr_ct = lang.ilen(
            a for a in (
                'request_cls',
                'option_cls_set',
                'new_cls',
                'response_cls',
            ) if hasattr(cls, a)
        )
        if present_attr_ct:
            if present_attr_ct != 4:
                raise AttributeError('Must set all service attrs if any set')
            return

        request_ann, option_ann, new_ann, response_ann = service_base.args

        request_cls: type[ServiceRequest] = check.issubclass(check.isinstance(request_ann, type), ServiceRequest)  # noqa
        response_cls: type[ServiceResponse] = check.issubclass(check.isinstance(response_ann, type), ServiceResponse)  # noqa
        new_cls: ta.Any = new_ann

        option_cls_set: frozenset[type[Option]]
        if isinstance(option_ann, rfl.Union):
            option_cls_set = frozenset(check.issubclass(check.isinstance(a, type), Option) for a in option_ann.args)  # noqa
        else:
            option_cls_set = frozenset([check.issubclass(check.isinstance(option_ann, type), Option)])  # noqa

        cls.request_cls = request_cls
        cls.option_cls_set = option_cls_set
        cls.new_cls = new_cls
        cls.response_cls = response_cls

    @abc.abstractmethod
    def invoke(self, request: ServiceRequestT) -> ServiceResponseT:
        raise NotImplementedError

    @ta.final
    def invoke_new(
            self,
            v: ServiceNewT,
            *options: ServiceOptionT,
            **kwargs: ta.Any,
    ) -> ServiceResponseT:
        request = self.request_cls.new(
            v,
            *options,
            **kwargs,
        )
        return self.invoke(ta.cast(ServiceRequestT, request))

    def __call__(
            self,
            v: ServiceNewT,
            *options: ServiceOptionT,
            **kwargs: ta.Any,
    ) -> ServiceResponseT:
        return self.invoke_new(v, *options, **kwargs)



import abc
import enum
import typing as ta

from omlish import dataclasses as dc
from omlish import lang

from .options import Option
from .services import Service
from .services import ServiceOption
from .services import ServiceRequest
from .services import ServiceResponse


ModelInputT = ta.TypeVar('ModelInputT')
ModelOptionT = ta.TypeVar('ModelOptionT', bound='Option')
ModelNewT = ta.TypeVar('ModelNewT')
ModelOutputT = ta.TypeVar('ModelOutputT')

ModelRequestT = ta.TypeVar('ModelRequestT', bound='ModelRequest')
ModelResponseT = ta.TypeVar('ModelResponseT', bound='ModelResponse')


##


class FinishReason(enum.Enum):
    STOP = enum.auto()
    LENGTH = enum.auto()
    TOOL_EXEC = enum.auto()
    CONTENT_FILTER = enum.auto()
    OTHER = enum.auto()


@dc.dataclass(frozen=True)
class TokenUsage(lang.Final):
    input: int
    output: int
    total: int


##


class ModelOption(Option, lang.Abstract):
    pass


##


ModelOptions: ta.TypeAlias = ServiceOption | ModelOption


@dc.dataclass(frozen=True, kw_only=True)
class ModelRequest(
    ServiceRequest[
        ModelInputT,
        ModelOptionT,
        ModelNewT,
    ],
    lang.Abstract,
):
    pass


@dc.dataclass(frozen=True, kw_only=True)
class ModelResponse(ServiceResponse[ModelOutputT], lang.Abstract):
    usage: TokenUsage | None = dc.xfield(None, repr_fn=dc.opt_repr)
    reason: FinishReason | None = dc.xfield(None, repr_fn=dc.opt_repr)


class Model(
    Service[
        ModelRequestT,
        ModelOptionT,
        ModelNewT,
        ModelResponseT,
    ],
    lang.Abstract,
):
    @abc.abstractmethod
    def invoke(self, request: ModelRequestT) -> ModelResponseT:
        raise NotImplementedError



import abc
import collections.abc
import typing as ta

from omlish import dataclasses as dc
from omlish import dispatch

from .chat.messages import AiMessage
from .chat.messages import Message
from .chat.messages import SystemMessage
from .chat.messages import ToolExecResultMessage
from .chat.messages import UserMessage
from .content.content import Content
from .content.images import Image
from .content.placeholders import Placeholder
from .services import ServiceRequest


StringTransformable = ta.Union[  # noqa
    str,
    ta.Sequence['StringTransformable'],
    Content,
    Message,
    ServiceRequest,
]

StringTransformableT = ta.TypeVar('StringTransformableT', bound=StringTransformable)


class StringTransform:
    @dispatch.method
    def apply(self, s: StringTransformableT) -> StringTransformableT:
        raise TypeError(s)

    #

    @apply.register
    @abc.abstractmethod
    def apply_str(self, s: str) -> str:
        raise NotImplementedError

    @apply.register
    def apply_sequence(self, s: collections.abc.Sequence) -> collections.abc.Sequence:
        return [self.apply(e) for e in s]

    #

    @apply.register
    def apply_image(self, s: Image) -> Image:
        return s

    @apply.register
    def apply_placeholder(self, s: Placeholder) -> Placeholder:
        return s

    #

    @apply.register
    def apply_ai_message(self, s: AiMessage) -> AiMessage:
        return s

    @apply.register
    def apply_system_message(self, s: SystemMessage) -> SystemMessage:
        return dc.replace(s, s=self.apply(s.s))

    @apply.register
    def apply_tool_exec_result_message(self, s: ToolExecResultMessage) -> ToolExecResultMessage:
        return s

    @apply.register
    def apply_user_message(self, s: UserMessage) -> UserMessage:
        return dc.replace(s, content=self.apply(s.c))

    #

    @apply.register
    def apply_service_request(self, s: ServiceRequest) -> ServiceRequest:
        return dc.replace(s, v=self.apply(s.v))


@dc.dataclass(frozen=True)
class FnStringTransform(StringTransform):
    fn: ta.Callable[[str], str]

    @StringTransform.apply.register  # noqa
    def apply_str(self, s: str) -> str:
        return self.fn(s)


def transform_strings(fn: ta.Callable[[str], str], s: StringTransformableT) -> StringTransformableT:
    return FnStringTransform(fn).apply(s)



"""
https://docs.anthropic.com/en/docs/about-claude/models#model-comparison-table

https://github.com/anthropics/anthropic-sdk-python/tree/cd80d46f7a223a5493565d155da31b898a4c6ee5/src/anthropic/types
https://github.com/anthropics/anthropic-sdk-python/blob/cd80d46f7a223a5493565d155da31b898a4c6ee5/src/anthropic/resources/completions.py#L53
https://github.com/anthropics/anthropic-sdk-python/blob/cd80d46f7a223a5493565d155da31b898a4c6ee5/src/anthropic/resources/messages.py#L70
"""
import os
import typing as ta

from omlish import check
from omlish.formats import json
from omlish.http import all as http
from omlish.secrets.secrets import Secret

from ...chat.messages import AiMessage
from ...chat.messages import Message
from ...chat.messages import SystemMessage
from ...chat.messages import UserMessage
from ...chat.models import AiChoice
from ...chat.models import ChatModel
from ...chat.models import ChatRequest
from ...chat.models import ChatResponse


class AnthropicChatModel(ChatModel):
    model: ta.ClassVar[str] = (
        'claude-3-5-sonnet-20241022'
        # 'claude-3-opus-20240229'
    )

    ROLES_MAP: ta.ClassVar[ta.Mapping[type[Message], str]] = {
        SystemMessage: 'system',
        UserMessage: 'user',
        AiMessage: 'assistant',
    }

    def __init__(
            self,
            *,
            api_key: Secret | str | None = None,
    ) -> None:
        super().__init__()
        self._api_key = Secret.of(api_key if api_key is not None else os.environ['ANTHROPIC_API_KEY'])

    def _get_msg_content(self, m: Message) -> str | None:
        if isinstance(m, (SystemMessage, AiMessage)):
            return m.s

        elif isinstance(m, UserMessage):
            return check.isinstance(m.c, str)

        else:
            raise TypeError(m)

    def invoke(
            self,
            request: ChatRequest,
            *,
            max_tokens: int = 4096,
    ) -> ChatResponse:
        messages = []
        system: str | None = None
        for i, m in enumerate(request.v):
            if isinstance(m, SystemMessage):
                if i != 0 or system is not None:
                    raise Exception('Only supports one system message and must be first')
                system = self._get_msg_content(m)
            else:
                messages.append(dict(
                    role=self.ROLES_MAP[type(m)],  # noqa
                    content=check.isinstance(self._get_msg_content(m), str),
                ))

        raw_request = dict(
            model=self.model,
            **(dict(system=system) if system is not None else {}),
            messages=messages,
            max_tokens=max_tokens,
        )

        raw_response = http.request(
            'https://api.anthropic.com/v1/messages',
            headers={
                http.consts.HEADER_CONTENT_TYPE: http.consts.CONTENT_TYPE_JSON,
                b'x-api-key': self._api_key.reveal().encode('utf-8'),
                b'anthropic-version': b'2023-06-01',
            },
            data=json.dumps(raw_request).encode('utf-8'),
        )

        response = json.loads(check.not_none(raw_response.data).decode('utf-8'))

        return ChatResponse(v=[
            AiChoice(AiMessage(response['content'][0]['text'])),  # noqa
        ])



"""
TODO:
 - linux clipboard
"""
import argparse
import io
import os.path
import sys
import typing as ta

from omdev.cli import CliModule
from omlish import check
from omlish import lang


if ta.TYPE_CHECKING:
    import pytesseract
    import rapidocr_onnxruntime as rapidocr
    from PIL import Image

    from omdev.clipboard import darwin_cf as darwin_clipboard

else:
    pytesseract = lang.proxy_import('pytesseract')
    rapidocr = lang.proxy_import('rapidocr_onnxruntime')
    Image = lang.proxy_import('PIL.Image')

    darwin_clipboard = lang.proxy_import('omdev.clipboard.darwin_cf')


##


Ocr: ta.TypeAlias = ta.Callable[['Image.Image'], str]

OCR_BACKENDS: ta.Mapping[str, Ocr] = {
    'rapidocr': lambda img: '\n'.join(text[1] for text in rapidocr.RapidOCR()(img)[0] or []),
    'tesseract': lambda img: pytesseract.image_to_string(img),
}

DEFAULT_OCR_BACKEND = 'rapidocr'


##


def _get_img_data(file: str | None) -> ta.Any:
    if file == '@':
        if sys.platform == 'darwin':
            cis = darwin_clipboard.get_darwin_clipboard_data(types={'public.png'})
            if not cis:
                raise RuntimeError('No clipboard image data found')
            return io.BytesIO(check.not_none(cis[0].data))

        else:
            raise OSError(sys.platform)

    elif file:
        return os.path.expanduser(file)

    else:
        return sys.stdin.buffer


def _main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('file', nargs='?')
    parser.add_argument('-b', '--backend', choices=list(OCR_BACKENDS), default=DEFAULT_OCR_BACKEND)
    args = parser.parse_args()

    #

    ocr = OCR_BACKENDS[args.backend]

    img_data = _get_img_data(args.file)

    with Image.open(img_data) as img:
        text = ocr(img)

    print(text)


# @omlish-manifest
_CLI_MODULE = CliModule('ocr', __name__)


if __name__ == '__main__':
    _main()



import enum
import functools
import typing as ta

from omlish import dataclasses as dc
from omlish import lang

from .docs import FieldRef


##


class Filter(lang.Abstract):
    @property
    def children(self) -> ta.Iterable['Filter']:
        return ()

    def __and__(self, other) -> 'Filter':
        return and_(self, other)

    def __or__(self, other) -> 'Filter':
        return or_(self, other)

    def __neg__(self) -> 'Filter':
        return not_(self)


##


class MultiOp(enum.Enum):
    AND = '&&'
    OR = '||'


@dc.dataclass(frozen=True)
class MultiFilter(Filter, lang.Final):
    op: MultiOp = dc.xfield(check_type=True)
    children: ta.Sequence[Filter] = dc.xfield(
        override=True,
        validate=lambda v: len(v) > 1 and all(isinstance(e, Filter) for e in v),
    )

    def __str__(self) -> str:
        return f'({f" {self.op.value} ".join(map(str, self.children))})'


def multi(op: MultiOp, *children: Filter) -> Filter:
    if len(children) == 1:
        return children[0]
    return MultiFilter(op, children)


and_ = functools.partial(multi, MultiOp.AND)
or_ = functools.partial(multi, MultiOp.OR)


##


@dc.dataclass(frozen=True)
class Not(Filter, lang.Final):
    child: Filter = dc.xfield(check_type=True)

    @property
    def children(self) -> ta.Iterable[Filter]:
        return (self.child,)


def not_(child: Filter) -> Filter:
    if isinstance(child, Not):
        return child.child
    return Not(child)


##


class CmpOp(enum.Enum):
    EQ = '='
    NE = '!='
    LT = '<'
    LE = '<='
    GT = '>'
    GE = '>='


@dc.dataclass(frozen=True)
class Cmp(Filter, lang.Final):
    op: CmpOp = dc.xfield(check_type=True)
    field: FieldRef = dc.xfield(check_type=True)
    value: ta.Any = dc.xfield()

    def __str__(self) -> str:
        return f'{self.field} {self.op.value} {self.value}'


def cmp(op: CmpOp, field: FieldRef, value: ta.Any) -> Filter:
    return Cmp(op, field, value)


eq = functools.partial(cmp, CmpOp.EQ)
ne = functools.partial(cmp, CmpOp.NE)
lt = functools.partial(cmp, CmpOp.LT)
le = functools.partial(cmp, CmpOp.LE)
gt = functools.partial(cmp, CmpOp.GT)
ge = functools.partial(cmp, CmpOp.GE)


def in_(field: FieldRef, *values: ta.Any) -> Filter:
    return or_(*[eq(field, v) for v in values])



import typing as ta

from omlish import lang

from .options import Option
from .services import Service
from .services import ServiceRequest
from .services import ServiceResponse


ServiceRequestT = ta.TypeVar('ServiceRequestT', bound='ServiceRequest')
ServiceOptionT = ta.TypeVar('ServiceOptionT', bound='Option')
ServiceNewT = ta.TypeVar('ServiceNewT')
ServiceResponseT = ta.TypeVar('ServiceResponseT', bound='ServiceResponse')


class WrapperService(
    Service[
        ServiceRequestT,
        ServiceOptionT,
        ServiceNewT,
        ServiceResponseT,
    ],
    lang.Abstract,
):
    def __init__(self, underlying: Service) -> None:
        super().__init__()
        self._underlying = underlying

    @property
    def underlying(self) -> Service[ServiceRequestT, ServiceOptionT, ServiceNewT, ServiceResponseT]:
        return self._underlying

    @property
    def request_cls(self) -> type[ServiceRequest]:  # type: ignore[override]
        return self._underlying.request_cls

    @property
    def option_cls_set(self) -> frozenset[type[Option]]:  # type: ignore[override]
        return self._underlying.option_cls_set

    @property
    def new_cls(self) -> ta.Any:
        return self._underlying.new_cls

    @property
    def response_cls(self) -> type[ServiceResponse]:  # type: ignore[override]
        return self._underlying.response_cls

    def invoke(self, request: ServiceRequestT) -> ServiceResponseT:
        return self._underlying.invoke(request)



import enum
import typing as ta

from omlish import check
from omlish import lang


if ta.TYPE_CHECKING:
    import numpy as np
else:
    np = lang.proxy_import('numpy')


Np: ta.TypeAlias = 'np.ndarray'


##


class Similarity(enum.Enum):
    DOT = enum.auto()
    COSINE = enum.auto()


CalcNpSimilaritiesFunc: ta.TypeAlias = ta.Callable[[Np, Np], Np]

CALC_NP_SIMILARITIES_FUNCS: ta.Mapping[Similarity, CalcNpSimilaritiesFunc] = {}


def register_calc_np_similarities_func(s):
    def inner(fn):
        check.not_in(s, CALC_NP_SIMILARITIES_FUNCS)
        ta.cast(dict, CALC_NP_SIMILARITIES_FUNCS)[s] = fn
        return fn
    return inner


def calc_np_similarities(
        similarity: Similarity,
        haystack: Np,  # (n, d)
        needle: Np,  # (d,)
) -> Np:
    return CALC_NP_SIMILARITIES_FUNCS[similarity](haystack, needle)


##


@register_calc_np_similarities_func(Similarity.DOT)
def calc_np_dot_similarities(
        haystack: Np,  # (n, d)
        needle: Np,  # (d,)
) -> Np:  # (n,)
    return np.dot(haystack, needle)


@register_calc_np_similarities_func(Similarity.COSINE)
def calc_np_cosine_similarities(
        haystack: Np,  # (n, d)
        needle: Np,  # (d,)
) -> Np:  # (n,)
    return np.dot(haystack, needle) / (np.linalg.norm(haystack, axis=1) * np.linalg.norm(needle))



import abc
import typing as ta

from omlish import dataclasses as dc
from omlish import lang

from .messages import Chat
from .messages import Message
from .models import ChatModel
from .models import ChatRequest
from .models import ChatResponse


class ChatHistory(lang.Abstract):
    @abc.abstractmethod
    def add(self, *msgs: Message) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def get(self) -> Chat:
        raise NotImplementedError

    @abc.abstractmethod
    def clear(self) -> None:
        raise NotImplementedError


class ListChatHistory(ChatHistory):
    def __init__(self, init: ta.Iterable[Message] | None = None) -> None:
        super().__init__()
        self._lst = list(init or ())

    def add(self, *msgs: Message) -> None:
        self._lst.extend(msgs)

    def get(self) -> Chat:
        return self._lst

    def clear(self) -> None:
        self._lst.clear()


class ChatHistoryModel(ChatModel):
    def __init__(
            self,
            underlying: ChatModel,
            history: ChatHistory,
    ) -> None:
        super().__init__()
        self._underlying = underlying
        self._history = history

    def invoke(self, request: ChatRequest) -> ChatResponse:
        new_req = dc.replace(request, v=[*self._history.get(), *request.v])
        response = self._underlying.invoke(new_req)
        self._history.add(
            *request.v,
            response.v[0].m,
        )
        return response



"""
bzip2 -cdk enwiki-20240801-pages-articles-multistream.xml.bz2 | lz4 -c > enwiki-20240801-pages-articles-multistream.xml.lz4

0 B / 42_781_970_578 B - 0.00 % - 0 B/s, 84404 elements, lang.is_gil_enabled()=False
"""  # noqa
import contextlib
import io
import os.path
import sys
import typing as ta

from omlish import lang
from omlish import marshal as msh
from omlish.formats import json

from . import models as mdl
from .utils import io as iou
from .utils import xml as xu
from .xml import parse_page


##


WIKI_REV = '20241101'


BZ2_INDEX_FILE_PATH = os.path.expanduser(f'~/data/enwiki/enwiki-{WIKI_REV}-pages-articles-multistream-index.txt.bz2')

#  23_851_879_117 compressed
# 103_090_295_026 uncompressed
BZ2_XML_FILE_PATH = os.path.expanduser(f'~/data/enwiki/enwiki-{WIKI_REV}-pages-articles-multistream.xml.bz2')

#  42_781_970_578 compressed
# 103_090_295_026 uncompressed
LZ4_XML_FILE_PATH = os.path.expanduser(f'~/data/enwiki/enwiki-{WIKI_REV}-pages-articles-multistream.xml.lz4')


def _main() -> None:
    # fp = BZ2_INDEX_FILE_PATH

    # fp = BZ2_XML_FILE_PATH
    fp = LZ4_XML_FILE_PATH

    use_subproc = False
    # use_subproc = True

    # use_lxml = False
    use_lxml = True

    output_dir = '.data/wiki/json'
    os.makedirs(output_dir, exist_ok=True)

    bs: ta.Any
    fpr: iou.FileProgressReporter | None
    with contextlib.ExitStack() as es:
        bs, fpr = es.enter_context(iou.open_compressed_reader(  # noqa
            fp,
            use_subprocess=use_subproc,
        ))

        mfw: iou.MultiFileWriter = es.enter_context(contextlib.closing(iou.MultiFileWriter(
            iou.Lz4MfwFile,
            os.path.join(output_dir, f'enwiki-{WIKI_REV}-pages-articles-%02d.jsonl.lz4'),
        )))

        if not use_lxml:
            cs = io.TextIOWrapper(bs, 'utf-8')
            it = xu.yield_root_children(cs)

        else:
            it = xu.yield_root_children(bs, use_lxml=True)

        root = next(it)  # noqa
        for i, el in enumerate(it):  # noqa
            if fpr is not None:
                if fpr.report():
                    print(f'{i} elements, {lang.is_gil_enabled()=}', file=sys.stderr)
            elif i and (i % 100_000) == 0:
                print(f'{i} elements, {lang.is_gil_enabled()=}', file=sys.stderr)

            if xu.strip_ns(el.tag) == 'page':
                page: mdl.Page = parse_page(el)

                oj = json.dumps_compact(msh.marshal(page))
                mfw.write(oj.encode('utf-8'), b'\n')


if __name__ == '__main__':
    _main()



import typing as ta

from omlish import lang

from ..generative import GenerativeOption
from ..models import ModelOption
from ..options import Option


class ChatOption(Option, lang.Abstract):
    pass


ChatOptions: ta.TypeAlias = ModelOption | GenerativeOption | ChatOption



"""
TODO:
 - lxml? faster?
  - xml.etree ~60 MiB/s INITIALLY, slows to ~30 MiB/s
  - lxml ~110 MiB/s INITIALLY, slows to ~50 MiB/s
  - https://lxml.de/1.3/compatibility.html
  - https://lxml.de/performance.html
  - https://lxml.de/apidoc/lxml.etree.html#lxml.etree.iterparse
 - https://docs.python.org/3/library/xml.etree.elementtree.html#xml.etree.ElementTree.XMLPullParser - sans-io
"""
import dataclasses as dc
import typing as ta
import xml.etree.ElementTree as ET

from omlish import lang
from omlish.formats.xml import strip_ns


if ta.TYPE_CHECKING:
    import lxml.etree as lxml_etree
else:
    lxml_etree = lang.proxy_import('lxml.etree')


T = ta.TypeVar('T')

Element: ta.TypeAlias = ta.Union[ET.Element, 'lxml_etree.Element']


##


ITER_PARSE_EVENTS = ('start', 'end', 'comment', 'pi', 'start-ns', 'end-ns')


def yield_root_children(
        source: ta.Any,
        *,
        retain_on_root: bool = False,
        use_lxml: bool = False,
        **kwargs: ta.Any,
) -> ta.Iterator[Element]:
    if use_lxml:
        parser = lxml_etree.iterparse
    else:
        parser = ET.iterparse
    it = iter(parser(source, ('start', 'end'), **kwargs))

    ev, root = next(it)
    if ev != 'start':
        raise RuntimeError(ev)
    yield root

    depth = 0
    for ev, el in it:
        if ev == 'start':
            depth += 1

        elif ev == 'end':
            depth -= 1
            if not depth:
                if not retain_on_root:
                    root.remove(el)
                yield el


##


@dc.dataclass(frozen=True, kw_only=True)
class ElementToKwargs:
    attrs: ta.Mapping[str, tuple[str, ta.Callable[[str], ta.Any]] | str | None] = dc.field(default_factory=dict)
    scalars: ta.Mapping[str, tuple[str, ta.Callable[[str | None], ta.Any]] | str | None] = dc.field(default_factory=dict)  # noqa
    single_children: ta.Mapping[str, tuple[str, ta.Callable[[Element], ta.Any]]] = dc.field(default_factory=dict)
    list_children: ta.Mapping[str, tuple[str, ta.Callable[[Element], ta.Any]]] = dc.field(default_factory=dict)
    text: str | None = None

    def __call__(self, el: Element) -> ta.Mapping[str, ta.Any]:
        kw: dict[str, ta.Any] = {}

        def set_kw(k: str, v: ta.Any) -> None:
            if k in kw:
                raise KeyError(k)
            kw[k] = v

        if el.attrib:
            for k, v in el.attrib.items():
                k = strip_ns(k)

                if k in self.attrs:
                    t = self.attrs[k]
                    if t is not None:
                        if isinstance(t, str):
                            set_kw(t, v)
                        else:
                            ak, fn = t
                            set_kw(ak, fn(v))

                else:
                    raise KeyError(k)

        for cel in el:
            k = strip_ns(cel.tag)

            if k in self.scalars:
                t = self.scalars[k]
                if t is not None:
                    if isinstance(t, str):
                        set_kw(t, cel.text)
                    else:
                        sk, sfn = t
                        set_kw(sk, sfn(cel.text))

            elif k in self.single_children:
                ck, fn = self.single_children[k]
                set_kw(ck, fn(cel))

            elif k in self.list_children:
                lk, fn = self.list_children[k]
                kw.setdefault(lk, []).append(fn(cel))

            else:
                raise KeyError(k)

        if self.text is not None:
            set_kw(self.text, el.text)

        return kw


@dc.dataclass(frozen=True)
class ElementToObj(ta.Generic[T]):
    cls: type[T]
    kw: ElementToKwargs

    def __call__(self, el: Element) -> T:
        return self.cls(**self.kw(el))



from .docs import (  # noqa
    Doc,
    DocField,
    DocSchema,
    FieldRef,
    f,
    f_,
)

from .filters import (  # noqa
    Cmp,
    CmpOp,
    Filter,
    MultiFilter,
    MultiOp,
    Not,
    and_,
    cmp,
    eq,
    ge,
    gt,
    in_,
    le,
    lt,
    multi,
    ne,
    not_,
    or_,
)

from .dtypes import (  # noqa
    Dtype,
    Dtypes,
)



import abc
import typing as ta

from .services import Service
from .services import ServiceRequest
from .services import ServiceResponse
from .strings import transform_strings
from .wrappers import WrapperService


##


class Templater(abc.ABC):
    @abc.abstractmethod
    def escape(self, s: str) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def apply(self, s: str) -> str:
        raise NotImplementedError


class DictTemplater(Templater):
    def __init__(
            self,
            env: ta.Mapping[str, str],
    ) -> None:
        super().__init__()
        self._env = env

    def escape(self, s: str) -> str:
        return s.replace('{', '{{').replace('}', '}}')

    def apply(self, s: str) -> str:
        return s.format(**self._env)


##


class TemplatingService(WrapperService):
    def __init__(self, underlying: Service, templater: Templater) -> None:
        super().__init__(underlying)
        self._templater = templater

    def invoke(self, request: ServiceRequest) -> ServiceResponse:
        out_request = transform_strings(self._templater.apply, request)
        return super().invoke(out_request)



import dataclasses as dc
import os.path
import typing as ta

from omlish import marshal as msh
from omlish.formats import json


T = ta.TypeVar('T')


##


STATE_VERSION = 0


@dc.dataclass(frozen=True)
class MarshaledState:
    version: int
    payload: ta.Any


#


def marshal_state(obj: ta.Any, ty: type | None = None, *, version: int = STATE_VERSION) -> ta.Any:
    ms = MarshaledState(
        version=version,
        payload=msh.marshal(obj, ty),
    )
    return msh.marshal(ms)


def save_state(file: str, obj: ta.Any, ty: type[T] | None, *, version: int = STATE_VERSION) -> bool:
    dct = marshal_state(obj, ty, version=version)
    data = json.dumps_pretty(dct)
    with open(file, 'w') as f:
        f.write(data)
    return True


#


def unmarshal_state(obj: ta.Any, ty: type[T] | None = None, *, version: int = STATE_VERSION) -> T | None:
    ms = msh.unmarshal(obj, MarshaledState)
    if ms.version < version:
        return None
    return msh.unmarshal(ms.payload, ty)  # type: ignore


def load_state(file: str, ty: type[T] | None, *, version: int = STATE_VERSION) -> T | None:
    if not os.path.isfile(file):
        return None
    with open(file) as f:
        data = f.read()
    dct = json.loads(data)
    return unmarshal_state(dct, ty, version=version)



import abc
import collections.abc
import typing as ta

from omlish import dataclasses as dc
from omlish import lang

from ..generative import Generative
from ..models import Model
from ..models import ModelRequest
from ..models import ModelResponse
from .messages import AiMessage
from .messages import Chat
from .messages import Message
from .options import ChatOptions


##


@dc.dataclass(frozen=True)
class AiChoice(lang.Final):
    m: AiMessage


AiChoices: ta.TypeAlias = ta.Sequence[AiChoice]


##


ChatInput: ta.TypeAlias = Chat
ChatNew: ta.TypeAlias = ta.Any
ChatOutput: ta.TypeAlias = AiChoices


@dc.dataclass(frozen=True, kw_only=True)
@dc.extra_params(generic_init=True)
class ChatRequest(
    ModelRequest[
        ChatInput,
        ChatOptions,
        ChatNew,
    ],
    lang.Final,
):
    @classmethod
    def new(
            cls,
            v: ChatNew,
            *options: ChatOptions,
            **kwargs: ta.Any,
    ) -> 'ChatRequest':
        if isinstance(v, Message):
            v = (v,)

        return super().new(
            v,
            *options,
            **kwargs,
        )

    @dc.validate
    def _validate_v(self) -> bool:
        return (
            not isinstance(self.v, str) and  # type: ignore[unreachable]
            isinstance(self.v, collections.abc.Sequence) and
            all(isinstance(e, Message) for e in self.v)
        )


@dc.dataclass(frozen=True, kw_only=True)
class ChatResponse(ModelResponse[ChatOutput], lang.Final):
    pass


class ChatModel(
    Model[
        ChatRequest,
        ChatOptions,
        ChatNew,
        ChatResponse,
    ],
    Generative,
    lang.Abstract,
):
    @abc.abstractmethod
    def invoke(self, request: ChatRequest) -> ChatResponse:
        raise NotImplementedError



import typing as ta

from omlish import collections as col
from omlish import dataclasses as dc
from omlish import lang

# from ..models import RequestContextItem
from .content import ExtendedContent


##


EnvKey = ta.NewType('EnvKey', str)


class Env(
    col.FrozenDict[str, str],
    # RequestContextItem,
    lang.Final,
):
    pass


##


@dc.dataclass(frozen=True)
class Placeholder(ExtendedContent, lang.Final):
    k: EnvKey
